import shutil

import pytest
from _pytest.fixtures import SubRequest

import gdsfactory as gf
from gdsfactory.config import CONFIG, diff_path

# from gdsfactory import clear_cache


@pytest.fixture(scope="session", autouse=False)
def cleandir(request: SubRequest) -> None:
    # clear_cache()
    build_folder = CONFIG["build_directory"]
    module_path = CONFIG["module_path"]

    if diff_path.exists():
        shutil.rmtree(diff_path)

    if build_folder.exists() and "noautofixt" not in request.keywords:
        shutil.rmtree(build_folder)

    for build_folder in module_path.glob("**/build"):
        shutil.rmtree(build_folder)


@pytest.fixture
def datadir(original_datadir, tmpdir):
    return original_datadir


@pytest.fixture(scope="session")
def show_diffs(request: SubRequest) -> None:
    c = gf.read.from_gdspaths(diff_path.glob("*.gds"))
    c.show(show_ports=True)


collect_ignore = ["difftest.py"]



from pytest_regressions.data_regression import DataRegressionFixture

from gdsfactory.config import PATH
from gdsfactory.klayout_tech import LayerDisplayProperties


def test_klayout_tech_create(
    data_regression: DataRegressionFixture, check: bool = True
) -> LayerDisplayProperties:

    lyp = LayerDisplayProperties.from_lyp(str(PATH.klayout_lyp))

    if check:
        data_regression.check(lyp.dict())

    return lyp


if __name__ == "__main__":
    lyp = test_klayout_tech_create(None, check=False)
    d = lyp.dict()



"""A GDS layer is a tuple of two integers.

You can:

- Load LayerColors from Klayout XML file (.lyp) (recommended)
- Define your layers in a Pydantic BaseModel

LayerColors based on phidl.device_layout
load_lyp, name_to_description, name_to_short_name based on phidl.utilities
preview_layerset based on phidl.geometry

"""
import pathlib
from functools import partial
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import numpy as np
import xmltodict
from pydantic import BaseModel, Field, validator

from gdsfactory.constants import _CSS3_NAMES_TO_HEX
from gdsfactory.name import clean_name

module_path = pathlib.Path(__file__).parent.absolute()
layer_path = module_path / "klayout" / "tech" / "layers.lyp"


def preview_layerset(ls, size: float = 100.0, spacing: float = 100.0) -> object:
    """Generates a Component with all the layers.

    Args:
        ls: LayerColors.
        size: square size.
        spacing: spacing between each square.

    """
    import gdsfactory as gf

    D = gf.Component(name="layerset", with_uuid=True)
    scale = size / 100
    num_layers = len(ls.layers)
    matrix_size = int(np.ceil(np.sqrt(num_layers)))
    sorted_layers = sorted(
        ls.layers.values(), key=lambda x: (x.gds_layer, x.gds_datatype)
    )
    for n, layer in enumerate(sorted_layers):
        gds_layer, gds_datatype = layer.gds_layer, layer.gds_datatype
        layer_tuple = (gds_layer, gds_datatype)
        R = gf.components.rectangle(size=(100 * scale, 100 * scale), layer=layer_tuple)
        T = gf.components.text(
            text=f"{layer.name}\n{layer.gds_layer} / {layer.gds_datatype}",
            size=20 * scale,
            position=(50 * scale, -20 * scale),
            justify="center",
            layer=layer_tuple,
        )

        xloc = n % matrix_size
        yloc = int(n // matrix_size)
        D.add_ref(R).movex((100 + spacing) * xloc * scale).movey(
            -(100 + spacing) * yloc * scale
        )
        D.add_ref(T).movex((100 + spacing) * xloc * scale).movey(
            -(100 + spacing) * yloc * scale
        )
    return D


class LayerColor(BaseModel):
    """Layer object with color, alpha (opacity) and dither.

    Parameters:
        gds_layer: int GDSII Layer number.
        gds_datatype: int GDSII datatype.
        name: str Name of the Layer.
        color: str Hex code of color for the Layer.
        alpha: Alpha parameter (opacity) for the Layer.
        dither: str KLayout dither parameter (texture) for the Layer
            (only used in phidl.utilities.write_lyp)

    """

    gds_layer: int = 0
    gds_datatype: int = 0
    name: str = "unnamed"
    description: Optional[str] = None
    inverted: bool = False
    color: Optional[str] = None
    alpha: float = 0.6
    dither: Optional[str] = None

    @validator("color")
    def color_is_valid(cls, color):
        try:
            if color is None:  # not specified
                color = None
            elif np.size(color) == 3:  # in format (0.5, 0.5, 0.5)
                color = np.array(color)
                if np.any(color > 1) or np.any(color < 0):
                    raise ValueError
                color = np.array(np.round(color * 255), dtype=int)
                color = "#{:02x}{:02x}{:02x}".format(*color)
            elif color[0] == "#":  # in format #1d2e3f
                if len(color) != 7:
                    raise ValueError
                int(color[1:], 16)  # Will throw error if not hex format
                color = color
            else:  # in named format 'gold'
                color = _CSS3_NAMES_TO_HEX[color.lower()]
        except Exception as error:
            raise ValueError(
                "LayerColor() color must be specified as a "
                "0-1 RGB triplet, (e.g. [0.5, 0.1, 0.9]), an HTML hex color string "
                "(e.g. '#a31df4'), or a CSS3 color name (e.g. 'gold' or "
                "see http://www.w3schools.com/colors/colors_names.asp )"
            ) from error
        return color


class LayerColors(BaseModel):
    """LayerColor dict.

    Parameters:
        layers: dict of LayerColors.

    """

    layers: Dict[str, LayerColor] = Field(default_factory=dict)

    def add_layer(
        self,
        name: str = "unnamed",
        gds_layer: int = 0,
        gds_datatype: int = 0,
        description: Optional[str] = None,
        color: Optional[str] = None,
        inverted: bool = False,
        alpha: float = 0.6,
        dither: Optional[str] = None,
    ) -> None:
        """Adds a layer to LayerColors object for nice colors.

        Args:
            name: Name of the Layer.
            gds_layer: GDSII Layer number.
            gds_datatype: GDSII datatype.
            description: Layer description.
            color: Hex code of color for the Layer.
            inverted: If true, inverts the Layer.
            alpha: layer opacity between 0 and 1 (0: invisible,  1: opaque).
            dither: KLayout dither style for phidl.utilities.write_lyp().

        """
        new_layer = LayerColor(
            gds_layer=gds_layer,
            gds_datatype=gds_datatype,
            name=name,
            description=description,
            inverted=inverted,
            color=color,
            alpha=alpha,
            dither=dither,
        )
        if name in self.layers:
            raise ValueError(
                f"Adding {name!r} already defined {list(self.layers.keys())}"
            )
        else:
            self.layers[name] = new_layer

    def __repr__(self):
        """Prints the number of Layers in the LayerColors object."""
        return (
            f"LayerColors ({len(self.layers)} layers total) \n"
            f"{list(self.layers.keys())}"
        )

    def get(self, name: str) -> LayerColor:
        """Returns Layer from name."""
        if name not in self.layers:
            raise ValueError(f"Layer {name!r} not in {list(self.layers.keys())}")
        else:
            return self.layers[name]

    def __getitem__(self, val):
        """Allows accessing to the layer names like ls['gold2'].

        Args:
            val: Layer name to access within the LayerColors.

        Returns:
            self.layers[val]: Accessed Layer in the LayerColors.

        """
        try:
            return self.layers[val]
        except Exception as error:
            raise ValueError(
                f"Layer {val!r} not in LayerColors {list(self.layers.keys())}"
            ) from error

    def get_from_tuple(self, layer_tuple: Tuple[int, int]) -> LayerColor:
        """Returns Layer from layer tuple (gds_layer, gds_datatype)."""
        tuple_to_name = {
            (v.gds_layer, v.gds_datatype): k for k, v in self.layers.items()
        }
        if layer_tuple not in tuple_to_name:
            raise ValueError(
                f"Layer color {layer_tuple} not in {list(tuple_to_name.keys())}"
            )

        name = tuple_to_name[layer_tuple]
        return self.layers[name]

    def get_layer_tuples(self):
        return {(layer.gds_layer, layer.gds_datatype) for layer in self.layers.values()}

    def clear(self) -> None:
        """Deletes all layers in the LayerColors."""
        self.layers = {}

    def preview(self):
        return preview_layerset(self)


def _name_to_short_name(name_str: str) -> str:
    """Maps the name entry of the lyp element to a name of the layer.

    i.e. the dictionary key used to access it.
    Default format of the lyp name are:

        - key - layer/datatype - description
        - key - description

    """
    if name_str is None:
        raise OSError(f"layer {name_str} has no name")
    fields = name_str.split("-")
    name = fields[0].split()[0].strip()
    return clean_name(name, remove_dots=True)


def _name_to_description(name_str) -> str:
    """Gets the description of the layer contained in the lyp name field.

    It is not strictly necessary to have a description. If none there, it returns ''.

    Default format of the lyp name are:

        - key - layer/datatype - description
        - key - description

    """
    if name_str is None:
        raise OSError(f"layer {name_str!r} has no name")
    fields = name_str.split()
    return " ".join(fields[1:]) if len(fields) > 1 else ""


def _add_layer(
    entry, lys: LayerColors, shorten_names: bool = True
) -> Optional[LayerColors]:
    """Adds a layer entry to layer colors and Returns a LayerColors updated dict.

    Returns None for invalid entries.

    Args:
        entry: layer entry.Entry is a dict of one element of 'properties'.
        lys: LayerColors map.
        shorten_names: if True takes the first part of the layer as its name.

    """
    info = entry["source"].split("@")[0]

    # skip layers without name or with */*
    if "'" in info or "*" in info:
        return None

    name = entry.get("name") or entry.get("source")
    if not name:
        return None

    infos = info.split("/")

    if len(infos) <= 1:
        return None

    gds_layer, gds_datatype = info.split("/")
    gds_layer = gds_layer.split()[-1]
    gds_datatype = gds_datatype.split()[-1]

    # print(entry.keys())
    # print(name, entry["xfill"], entry["fill-color"])
    # if entry["visible"] == "false" or entry["xfill"] == "false":

    name = _name_to_short_name(name) if shorten_names else name
    dither = entry["dither-pattern"]

    if ("visible" in entry.keys()) and (entry["visible"] == "false"):
        alpha = 0.0
    elif ("transparent" in entry.keys()) and (entry["transparent"] == "false"):
        alpha = 0.1 if dither == "I1" else 1.0
    else:
        alpha = 0.5

    settings = {
        "gds_layer": int(gds_layer),
        "gds_datatype": int(gds_datatype),
        "color": entry["fill-color"],
        "dither": dither,
        "name": name,
        "description": _name_to_description(name),
        "alpha": alpha,
    }

    lys.add_layer(**settings)
    return lys


def load_lyp(filepath: Path) -> LayerColors:
    """Returns a LayerColors object from a Klayout lyp file layer properties \
    file."""
    with open(filepath) as fx:
        lyp_dict = xmltodict.parse(fx.read(), process_namespaces=True)

    # lyp files have a top level that just has one dict: layer-properties
    # That has multiple children 'properties', each for a layer. So it gives a list.

    lyp_list = lyp_dict["layer-properties"]["properties"]
    if not isinstance(lyp_list, list):
        lyp_list = [lyp_list]

    lys = LayerColors()

    for entry in lyp_list:
        try:
            group_members = entry["group-members"]
        except KeyError:  # it is a real layer
            _add_layer(entry, lys)
        else:  # it is a group of other entries
            if not isinstance(group_members, list):
                group_members = [group_members]
            for member in group_members:
                try:
                    _add_layer(member, lys)
                except Exception:
                    _add_layer(member, lys, shorten_names=False)
    return lys


load_lyp_generic = partial(load_lyp, filepath=layer_path)


def lyp_to_dataclass(lyp_filepath: Union[str, Path], overwrite: bool = True) -> str:
    """Returns python LayerMap script from a klayout layer properties file lyp."""
    filepathin = pathlib.Path(lyp_filepath)
    filepathout = filepathin.with_suffix(".py")

    if filepathout.exists() and not overwrite:
        raise FileExistsError(f"You can delete {filepathout}")

    script = """
from pydantic import BaseModel
from gdsfactory.types import Layer


class LayerMap(BaseModel):
"""
    lys = load_lyp(filepathin)
    for layer_name, layer in sorted(lys.layers.items()):
        script += (
            f"    {layer_name}: Layer = ({layer.gds_layer}, {layer.gds_datatype})\n"
        )

    script += """
    class Config:
        frozen = True
        extra = "forbid"


LAYER = LayerMap()
"""

    filepathout.write_text(script)
    return script


def test_load_lyp():
    from gdsfactory.config import layer_path

    lys = load_lyp(layer_path)
    assert len(lys.layers) > 10, len(lys.layers)
    return lys


try:
    LAYER_COLORS = load_lyp_generic()
except Exception:
    print(f"Error loading generic layermap in {layer_path!r}")
    LAYER_COLORS = LayerColors()


if __name__ == "__main__":
    # LAYER_COLORS = load_lyp_generic()
    # import gdsfactory as gf

    # c = gf.components.rectangle(layer=(123, 0))
    # c.plot()

    # print(LAYER_COLORS)
    # print(LAYER_STACK.get_from_tuple((1, 0)))
    # print(LAYER_STACK.get_layer_to_material())
    # layer = LayerColor(color="gold")
    # print(layer)

    # lys = test_load_lyp()
    c = preview_layerset(LAYER_COLORS)
    c.show(show_ports=True)
    # print(LAYERS_OPTICAL)
    # print(layer("wgcore"))
    # print(layer("wgclad"))
    # print(layer("padding"))
    # print(layer("TEXT"))
    # print(type(layer("wgcore")))



from typing import Optional

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.types import LayerSpec


@gf.cell
def ramp(
    length: float = 10.0,
    width1: float = 5.0,
    width2: Optional[float] = 8.0,
    layer: LayerSpec = "WG",
) -> Component:
    """Return a ramp component.

    Based on phidl.

    Args:
        length: Length of the ramp section.
        width1: Width of the start of the ramp section.
        width2: Width of the end of the ramp section (defaults to width1).
        layer: Specific layer to put polygon geometry on.
    """
    if width2 is None:
        width2 = width1
    xpts = [0, length, length, 0]
    ypts = [width1, width2, 0, 0]
    c = Component()
    c.add_polygon([xpts, ypts], layer=layer)
    c.add_port(
        name="o1", center=[0, width1 / 2], width=width1, orientation=180, layer=layer
    )
    c.add_port(
        name="o2",
        center=[length, width2 / 2],
        width=width2,
        orientation=0,
        layer=layer,
    )
    return c


if __name__ == "__main__":
    c = ramp()
    c.show(show_ports=True)



from typing import Dict, Optional, Tuple

import numpy as np
import pygmsh

from gdsfactory.pdk import get_layer_stack
from gdsfactory.tech import LayerStack
from gdsfactory.types import ComponentOrReference, Layer


def mesh3D(
    component: ComponentOrReference,
    base_resolution: float = 0.2,
    refine_resolution: Optional[Dict[Layer, float]] = None,
    padding: Tuple[float, float, float, float, float, float] = (
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
    ),
    layer_stack: Optional[LayerStack] = None,
    exclude_layers: Optional[Tuple[Layer, ...]] = None,
):
    """Returns gmsh 3D geometry of component.

    Similar to component.to_3d(), but returns a gmsh mesh with:
    - layer-dependent mesh resolution
    - different physical blocks for different objects
    - sub-object labels for introduction in physical solvers (e.g. edges)

    Args:
        component: Component or ComponentReference.
        base_resolution: background mesh resolution (um).
        refine_resolution: feature mesh resolution (um); layer dependent via a dict (default to base_resolution).
        padding: amount (west, east, south, north, down, up) to enlarge simulation region beyond features (um).

    """
    layer_stack = layer_stack or get_layer_stack()
    layer_to_thickness = layer_stack.get_layer_to_thickness()
    layer_to_zmin = layer_stack.get_layer_to_zmin()
    exclude_layers = exclude_layers or ()

    geometry = pygmsh.occ.geometry.Geometry()
    geometry.characteristic_length_min = base_resolution
    geometry.characteristic_length_max = base_resolution

    model = geometry.__enter__()

    zmin_cell = np.inf
    zmax_cell = -np.inf
    xmin_cell = np.inf
    ymin_cell = np.inf
    xmax_cell = -np.inf
    ymax_cell = -np.inf

    # Create element resolution dict
    refine_dict = {
        layer: refine_resolution[layer]
        if layer in refine_resolution.keys()
        else base_resolution
        for layer in component.get_layers()
    }

    # Features
    all_blocks = []
    for layer, polygons in component.get_polygons(by_spec=True).items():
        if (
            layer not in exclude_layers
            and layer in layer_to_thickness
            and layer in layer_to_zmin
        ):
            height = layer_to_thickness[layer]
            zmin_layer = layer_to_zmin[layer]
            zmax_layer = zmin_layer + height

            if zmin_layer < zmin_cell:
                zmin_cell = zmin_layer
            if zmax_layer > zmax_cell:
                zmax_cell = zmax_layer

            layer_blocks = []
            for i, polygon in enumerate(polygons):
                xmin_block = np.min(polygon[:, 0])
                xmax_block = np.max(polygon[:, 0])
                ymin_block = np.min(polygon[:, 1])
                ymax_block = np.max(polygon[:, 1])
                if xmin_block < xmin_cell:
                    xmin_cell = xmin_block
                if xmax_block > xmax_cell:
                    xmax_cell = xmax_block
                if ymin_block < ymin_cell:
                    ymin_cell = ymin_block
                if ymax_block > ymax_cell:
                    ymax_cell = ymax_block
                points = [
                    model.add_point(
                        [polygon_point[0], polygon_point[1], zmin_layer],
                        mesh_size=refine_dict[layer],
                    )
                    for polygon_point in polygon
                ]
                polygon_lines = [
                    model.add_line(points[i], points[i + 1])
                    for i in range(-1, len(points) - 1)
                ]
                polygon_loop = model.add_curve_loop(polygon_lines)
                polygon_surface = model.add_plane_surface(polygon_loop)
                polygon_top, polygon_volume, polygon_lat = model.extrude(
                    polygon_surface,
                    [0, 0, height],
                    num_layers=int(height / refine_dict[layer]),
                )
                model.add_physical(polygon_volume, f"{layer}_{i}")
                layer_blocks.append(polygon_volume)
            # Recursively compute boolean fragments to eliminate overlaps
            for block_index, block in enumerate(layer_blocks):
                if block_index == 0:
                    continue
                else:
                    block = model.boolean_fragments(
                        block,
                        layer_blocks[block_index - 1],
                        delete_first=True,
                        delete_other=True,
                    )

            all_blocks.append(block)

    xmin_cell -= padding[0]
    xmax_cell += padding[1]
    ymin_cell -= padding[2]
    ymax_cell += padding[3]
    zmin_cell -= padding[4]
    zmax_cell += padding[5]

    # Background oxide
    volume = model.add_box(
        [xmin_cell, ymin_cell, zmin_cell],
        [xmax_cell - xmin_cell, ymax_cell - ymin_cell, zmax_cell - zmin_cell],
        mesh_size=base_resolution,
    )
    for block in all_blocks:
        volume = model.boolean_fragments(volume, block)

    model.add_physical(volume, "oxide")
    geometry.generate_mesh(dim=3, verbose=True)

    return geometry


if __name__ == "__main__":

    import gdsfactory as gf

    heater1 = gf.components.straight_heater_metal(length=2)
    heater2 = gf.components.straight_heater_metal(length=2).move([0, 20])

    heaters = gf.Component()
    heaters << heater1
    heaters << heater2
    heaters.show()

    layers_to_keep = [(1, 0), (47, 0)]

    geometry = mesh3D(
        heaters,
        exclude_layers=[
            layer_name
            for layer_name in heaters.get_layers()
            if layer_name not in layers_to_keep
        ],
        refine_resolution={(1, 0): 0.05, (47, 0): 0.2},
        base_resolution=1,
        padding=[2, 2, 2, 2, 2, 2],
    )

    import gmsh

    gmsh.write("mesh.msh")
    gmsh.clear()
    geometry.__exit__()

    import meshio

    mesh_from_file = meshio.read("mesh.msh")

    def create_mesh(mesh, cell_type, prune_z=False):
        cells = mesh.get_cells_type(cell_type)
        cell_data = mesh.get_cell_data("gmsh:physical", cell_type)
        points = mesh.points
        return meshio.Mesh(
            points=points,
            cells={cell_type: cells},
            cell_data={"name_to_read": [cell_data]},
        )

    # line_mesh = create_mesh(mesh_from_file, "line", prune_z=True)
    # meshio.write("facet_mesh.xdmf", line_mesh)

    triangle_mesh = create_mesh(mesh_from_file, "tetra", prune_z=False)
    meshio.write("mesh.xdmf", triangle_mesh)

    # for layer, polygons in heaters.get_polygons(by_spec=True).items():
    #     print(layer, polygons)



import pathlib
import tempfile
from typing import Dict, Optional, Union

import meep as mp
import pydantic
from meep import mpb

import gdsfactory as gf
from gdsfactory.simulation.gmeep.get_material import get_material
from gdsfactory.simulation.gmeep.get_meep_geometry import (
    get_meep_geometry_from_cross_section,
)
from gdsfactory.types import CrossSectionSpec, LayerStack

mpb.Verbosity(0)

tmp = pathlib.Path(tempfile.TemporaryDirectory().name).parent / "meep"
tmp.mkdir(exist_ok=True)


@pydantic.validate_arguments
def get_mode_solver_cross_section(
    cross_section: CrossSectionSpec = "strip",
    sy: float = 2.0,
    sz: float = 2.0,
    resolution: int = 32,
    nmodes: int = 4,
    wavelength: float = 1.55,
    clad_mat: str = "sio2",
    layer_stack: Optional[LayerStack] = None,
    dispersive: bool = False,
    material_name_to_meep: Optional[Dict[str, Union[str, float]]] = None,
    **kwargs,
) -> mpb.ModeSolver:
    """Returns a mode_solver simulation.

    Args:
        cross_section: CrossSection to solve
        sy: simulation region width (um)
        sz: simulation region height (um)
        resolution: resolution (pixels/um)
        nmodes: number of modes
        wavelength: wavelength at which to compute mode (um)
        clad_mat: cladding material around CrossSection
        layer_stack: contains layer to thickness, zmin and material.
            Defaults to active pdk.layer_stack.
        dispersive: use dispersive material models (requires higher resolution).
        material_name_to_meep: dispersive materials have a wavelength
            dependent index. Maps layer_stack names with meep material database names.

    """
    x = gf.get_cross_section(cross_section=cross_section, **kwargs)

    # Define the computational cell.  We'll make x the propagation direction.
    # the other cell sizes should be big enough so that the boundaries are
    # far away from the mode field.
    geometry_lattice = mp.Lattice(size=mp.Vector3(0, sy, sz))

    geometry = get_meep_geometry_from_cross_section(
        cross_section=x,
        layer_stack=layer_stack,
        material_name_to_meep=material_name_to_meep,
        wavelength=wavelength,
        dispersive=dispersive,
    )

    # The k (i.e. beta, i.e. propagation constant) points to look at, in
    # units of 2*pi/um.  We'll look at num_k points from k_min to k_max.
    num_k = 9
    k_min = 0.1
    k_max = 3.0
    k_points = mp.interpolate(num_k, [mp.Vector3(k_min), mp.Vector3(k_max)])

    # Increase this to see more modes.  (The guided ones are the ones below the
    # light line, i.e. those with frequencies < kmag / 1.45, where kmag
    # is the corresponding column in the output if you grep for "freqs:".)
    # use this prefix for output files

    filename_prefix = tmp / f"cross_section_{x.name}"

    clad = get_material(
        name=clad_mat,
        wavelength=wavelength,
        material_name_to_meep=material_name_to_meep,
        dispersive=dispersive,
    )

    mode_solver = mpb.ModeSolver(
        geometry_lattice=geometry_lattice,
        geometry=geometry,
        k_points=k_points,
        resolution=resolution,
        num_bands=nmodes,
        default_material=clad,
        filename_prefix=str(filename_prefix),
        ensure_periodicity=False,
    )
    mode_solver.nmodes = nmodes
    mode_solver.info = dict(
        dispersive=dispersive,
        material_indices=[
            b.material.epsilon(1 / wavelength)[0, 0] ** 0.5 for b in geometry
        ],
        sy=sy,
        sz=sz,
        resolution=resolution,
        nmodes=nmodes,
    )
    return mode_solver


get_mode_solver_rib = gf.partial(get_mode_solver_cross_section, cross_section="rib")


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    cross_section = "pn"
    s = gf.components.straight(cross_section=cross_section)

    s.show()

    m = get_mode_solver_cross_section(
        cross_section=cross_section,
        resolution=200,
    )

    m.init_params(p=mp.NO_PARITY, reset_fields=False)
    eps = m.get_epsilon()
    cmap = "binary"
    origin = "lower"
    plt.imshow(
        eps.T**0.5,
        cmap=cmap,
        origin=origin,
        aspect="auto",
        extent=[
            -m.info["sy"] / 2,
            m.info["sy"] / 2,
            -m.info["sz"] / 2,
            m.info["sz"] / 2,
        ],
    )
    plt.colorbar()
    plt.show()



"""Based on phidl.geometry."""
from typing import Tuple

import gdspy
import numpy as np
from gdspy import clipper

import gdsfactory as gf
from gdsfactory.component_layout import Polygon, _parse_layer
from gdsfactory.types import Component, ComponentReference, LayerSpec


def _merge_nearby_floating_points(x, tol=1e-10):
    """Takes an array `x` and merges any values within the tolerance `tol`.

    Args:
        x: list of int or float
            Array of values with floating point errors.
        tol : float
            Tolerance within which points will be merged.

    Returns
        xsort : list of int or float Corrected and sorted array.

    Examples
    --------
    If given:
    >>> x = [-2, -1, 0, 1.0001, 1.0002, 1.0003, 4, 5, 5.003, 6, 7, 8]
    >>> _merge_nearby_floating_points(x, tol = 1e-3)
    Will return:
    >>> [-2, -1, 0, 1.0001, 1.0001, 1.0001, 4, 5, 5.003, 6, 7, 8].
    """
    xargsort = np.argsort(x)
    xargunsort = np.argsort(xargsort)
    xsort = x[xargsort]
    xsortthreshold = np.diff(xsort) < tol
    xsortthresholdind = np.argwhere(xsortthreshold)

    # Merge nearby floating point values
    for xi in xsortthresholdind:
        xsort[xi + 1] = xsort[xi]
    return xsort[xargunsort]


def _merge_floating_point_errors(polygons, tol=1e-10):
    """Fixes floating point errors in the input polygons by merging values \
    within `tol` tolerance. See _merge_nearby_floating_points for specifics.

    Args:
        polygons: PolygonSet or list of polygons
            Set of polygons with floating point errors.
        tol: float
            Tolerance within which points will be merged.

    Returns
        polygons_fixed : PolygonSet Set of corrected polygons.
    """
    stacked_polygons = np.vstack(polygons)
    x = stacked_polygons[:, 0]
    y = stacked_polygons[:, 1]
    polygon_indices = np.cumsum([len(p) for p in polygons])

    xfixed = _merge_nearby_floating_points(x, tol=tol)
    yfixed = _merge_nearby_floating_points(y, tol=tol)
    stacked_polygons_fixed = np.vstack([xfixed, yfixed]).T
    return np.vsplit(stacked_polygons_fixed, polygon_indices[:-1])


def _crop_region(polygons, left, bottom, right, top, precision):
    """Given a rectangular boundary defined by left/bottom/right/top, this \
    takes a list of polygons and cuts them at the boundary, discarding parts \
    of the polygons outside the rectangle.

    Args:
        polygons : PolygonSet or list of polygons
            Set or list of polygons to be cropped.
        left : int or float
            The x-coordinate of the lefthand boundary.
        bottom : int or float
            The y-coordinate of the bottom boundary.
        right : int or float
            The x-coordinate of the righthand boundary.
        top : int or float
            The y-coordinate of the top boundary.
        precision : float
            Desired precision for rounding vertex coordinates.

    Returns:
        cropped_polygons : PolygonSet or list of polygons
            Set or list of polygons that are cropped according to the specified
            boundary.
    """
    cropped_polygons = []
    for p in polygons:
        clipped_polys = clipper._chop(p, [top, bottom], 1, 1 / precision)
        # polygon, [cuts], axis, scale
        for cp in clipped_polys[1]:
            result = clipper._chop(cp, [left, right], 0, 1 / precision)
            cropped_polygons += list(result[1])
    return cropped_polygons


def _crop_edge_polygons(all_polygons, bboxes, left, bottom, right, top, precision):
    """Parses out which polygons are along the edge of the rectangle and need \
    to be cropped and which are deep inside the rectangle region and can be \
    left alone, then crops only those polygons along the edge.

    Args:
        all_polygons : PolygonSet or list of polygons
            Set or list of polygons to be cropped.
        bboxes : list
            List of all polygon bboxes in all_polygons.
        left : int or float
            The x-coordinate of the lefthand boundary.
        bottom : int or float
            The y-coordinate of the bottom boundary.
        right : int or float
            The x-coordinate of the righthand boundary.
        top : int or float
            The y-coordinate of the top boundary.
        precision : float
            Desired precision for rounding vertex coordinates.

    Returns:
        polygons_to_process : PolygonSet or list of polygons
            Set or list of polygons with crop applied to edge polygons.
    """
    polygons_in_rect_i = _find_bboxes_in_rect(bboxes, left, bottom, right, top)
    polygons_edge_i = _find_bboxes_on_rect_edge(bboxes, left, bottom, right, top)
    polygons_in_rect_no_edge_i = polygons_in_rect_i & (~polygons_edge_i)

    # Crop polygons along the edge and recombine them with polygons inside the
    # rectangle
    polygons_edge = all_polygons[polygons_edge_i]
    polygons_in_rect_no_edge = all_polygons[polygons_in_rect_no_edge_i].tolist()
    polygons_edge_cropped = _crop_region(
        polygons_edge, left, bottom, right, top, precision=precision
    )
    return polygons_in_rect_no_edge + polygons_edge_cropped


def _find_bboxes_in_rect(bboxes, left, bottom, right, top):
    """Given a list of polygon bounding boxes and a rectangle defined by \
    left/bottom/right/top, this function returns those polygons which overlap \
    the rectangle.

    Args:
        bboxes : list
            List of all polygon bboxes.
        left : int or float
            The x-coordinate of the lefthand boundary.
        bottom : int or float
            The y-coordinate of the bottom boundary.
        right : int or float
            The x-coordinate of the righthand boundary.
        top : int or float
            The y-coordinate of the top boundary.

    Returns:
        result : list
            List of all polygon bboxes that overlap with the defined rectangle.
    """
    return (
        (bboxes[:, 0] <= right)
        & (bboxes[:, 2] >= left)
        & (bboxes[:, 1] <= top)
        & (bboxes[:, 3] >= bottom)
    )


# _find_bboxes_on_rect_edge
def _find_bboxes_on_rect_edge(bboxes, left, bottom, right, top):
    """Given a list of polygon bounding boxes and a rectangular boundary \
    defined by left/bottom/right/top, this function returns those polygons \
    which intersect the rectangular boundary.

    Args:
        bboxes : list
            List of all polygon bboxes.
        left : int or float
            The x-coordinate of the lefthand boundary.
        bottom : int or float
            The y-coordinate of the bottom boundary.
        right : int or float
            The x-coordinate of the righthand boundary.
        top : int or float
            The y-coordinate of the top boundary.

    Returns:
        result: List of all polygon bboxes that intersect the defined
            rectangular boundary.
    """
    bboxes_left = _find_bboxes_in_rect(bboxes, left, bottom, left, top)
    bboxes_right = _find_bboxes_in_rect(bboxes, right, bottom, right, top)
    bboxes_top = _find_bboxes_in_rect(bboxes, left, top, right, top)
    bboxes_bottom = _find_bboxes_in_rect(bboxes, left, bottom, right, bottom)
    return bboxes_left | bboxes_right | bboxes_top | bboxes_bottom


def _offset_region(
    all_polygons,
    bboxes,
    left,
    bottom,
    right,
    top,
    distance=5,
    join_first=True,
    precision=1e-4,
    join="miter",
    tolerance=2,
):
    """Taking a region of e.g. size (x, y) which needs to be offset by \
    distance d, this function crops out a region (x+2*d, y+2*d) large, offsets \
    that region, then crops it back to size (x, y) to create a valid result.

    Args:
        all_polygons : PolygonSet or list of polygons
            Set or list of polygons to be cropped and offset.
        bboxes : list
            List of all polygon bboxes in all_polygons.
        left : int or float
            The x-coordinate of the lefthand boundary.
        bottom : int or float
            The y-coordinate of the bottom boundary.
        right : int or float
            The x-coordinate of the righthand boundary.
        top : int or float
            The y-coordinate of the top boundary.
        distance : int or float
            Distance to offset polygons. Positive values expand, negative shrink.
        join_first : bool
            Join all paths before offsetting to avoid unnecessary joins in
            adjacent polygon sides.
        precision : float
            Desired precision for rounding vertex coordinates.
        join : {'miter', 'bevel', 'round'}
            Type of join used to create the offset polygon.
        tolerance : int or float
            For miter joints, this number must be at least 2 and it represents the
            maximal distance in multiples of offset between new vertices and their
            original position before beveling to avoid spikes at acute joints. For
            round joints, it indicates the curvature resolution in number of
            points per full circle.

    Returns:
        polygons_offset_cropped:
            The resulting input polygons that are cropped to be between the
            coordinates (left, bottom, right, top)

    """
    # Mark out a region slightly larger than the final desired region
    d = distance * 1.01

    polygons_to_offset = _crop_edge_polygons(
        all_polygons,
        bboxes,
        left - d,
        bottom - d,
        right + d,
        top + d,
        precision=precision,
    )

    # Offset the resulting cropped polygons and recrop to final desired size
    polygons_offset = clipper.offset(
        polygons_to_offset, distance, join, tolerance, 1 / precision, int(join_first)
    )
    return _crop_region(polygons_offset, left, bottom, right, top, precision=precision)


def _polygons_to_bboxes(polygons):
    """Generates the bboxes of all input polygons.

    Args:
        polygons : PolygonSet or list of polygons
            Set or list of polygons to generate bboxes of.

    Returns:
        bboxes : list
            List of all polygon bboxes in polygons.
    """
    #    Build bounding boxes
    bboxes = np.empty([len(polygons), 4])
    for n, p in enumerate(polygons):
        left, bottom = np.min(p, axis=0)
        right, top = np.max(p, axis=0)
        bboxes[n] = [left, bottom, right, top]
    return bboxes


def _offset_polygons_parallel(
    polygons,
    distance=5,
    num_divisions=(10, 10),
    join_first=True,
    precision=1e-4,
    join="miter",
    tolerance=2,
):
    """Offsets a list of subsections and returns the offset polygons.

    Args:
        polygons : PolygonSet or list of polygons
        distance : int or float
            Distance to offset polygons. Positive values expand, negative shrink.
        num_divisions : array-like[2] of int
            The number of divisions with which the geometry is divided into
            multiple rectangular regions. This allows for each region to be
            processed sequentially, which is more computationally efficient.
        join_first : bool
            Join all paths before offsetting to avoid unnecessary joins in
            adjacent polygon sides.
        precision : float
            Desired precision for rounding vertex coordinates.
        join : {'miter', 'bevel', 'round'}
            Type of join used to create the offset polygon.
        tolerance : int or float
            For miter joints, this number must be at least 2 and it represents the
            maximal distance in multiples of offset between new vertices and their
            original position before beveling to avoid spikes at acute joints. For
            round joints, it indicates the curvature resolution in number of
            points per full circle.

    """
    # Build bounding boxes
    polygons = np.asarray(polygons)
    bboxes = _polygons_to_bboxes(polygons)

    xmin, ymin = np.min(bboxes[:, 0:2], axis=0) - distance
    xmax, ymax = np.max(bboxes[:, 2:4], axis=0) + distance

    xsize = xmax - xmin
    ysize = ymax - ymin
    xdelta = xsize / num_divisions[0]
    ydelta = ysize / num_divisions[1]
    xcorners = xmin + np.arange(num_divisions[0]) * xdelta
    ycorners = ymin + np.arange(num_divisions[1]) * ydelta

    offset_polygons = []
    for xc in xcorners:
        for yc in ycorners:
            left = xc
            right = xc + xdelta
            bottom = yc
            top = yc + ydelta
            _offset_region_polygons = _offset_region(
                polygons,
                bboxes,
                left,
                bottom,
                right,
                top,
                distance=distance,
                join_first=join_first,
                precision=precision,
                join=join,
                tolerance=tolerance,
            )
            offset_polygons += _offset_region_polygons

    return offset_polygons


@gf.cell
def offset(
    elements: Component,
    distance: float = 0.1,
    join_first: bool = True,
    precision: float = 1e-4,
    num_divisions: Tuple[int, int] = (1, 1),
    join: str = "miter",
    tolerance: int = 2,
    max_points: int = 4000,
    layer: LayerSpec = "WG",
) -> Component:
    """Returns an element containing all polygons with an offset Shrinks or \
    expands a polygon or set of polygons.

    Args:
        elements: Component(/Reference), list of Component(/Reference), or Polygon
          Polygons to offset or Component containing polygons to offset.
        distance: Distance to offset polygons. Positive values expand, negative shrink.
        precision: Desired precision for rounding vertex coordinates.
        num_divisions: The number of divisions with which the geometry is divided into
          multiple rectangular regions. This allows for each region to be
          processed sequentially, which is more computationally efficient.
        join: {'miter', 'bevel', 'round'} Type of join used to create polygon offset
        tolerance: For miter joints, this number must be at least 2 represents the
          maximal distance in multiples of offset between new vertices and their
          original position before beveling to avoid spikes at acute joints. For
          round joints, it indicates the curvature resolution in number of
          points per full circle.
        max_points: The maximum number of vertices within the resulting polygon.
        layer: Specific layer to put polygon geometry on.

    Returns:
        Component containing a polygon(s) with the specified offset applied.

    """
    if not isinstance(elements, list):
        elements = [elements]
    polygons_to_offset = []
    for e in elements:
        if isinstance(e, (Component, ComponentReference)):
            polygons_to_offset += e.get_polygons(by_spec=False)
        elif isinstance(e, (Polygon, gdspy.Polygon)):
            polygons_to_offset.append(e)
    if len(polygons_to_offset) == 0:
        return gf.Component("offset")
    polygons_to_offset = _merge_floating_point_errors(
        polygons_to_offset, tol=precision / 1000
    )

    layer = gf.get_layer(layer)
    gds_layer, gds_datatype = _parse_layer(layer)
    if all(np.array(num_divisions) == np.array([1, 1])):
        p = gdspy.offset(
            polygons_to_offset,
            distance=distance,
            join=join,
            tolerance=tolerance,
            precision=precision,
            join_first=join_first,
            max_points=max_points,
            layer=gds_layer,
            datatype=gds_datatype,
        )
    else:
        p = _offset_polygons_parallel(
            polygons_to_offset,
            distance=distance,
            num_divisions=num_divisions,
            join_first=join_first,
            precision=precision,
            join=join,
            tolerance=tolerance,
        )

    component = gf.Component()
    polygons = component.add_polygon(p, layer=layer)
    [
        polygon.fracture(max_points=max_points, precision=precision)
        for polygon in polygons
    ]
    return component


def test_offset() -> None:
    c = gf.components.ring()
    co = offset(c, distance=0.5)
    assert int(co.area()) == 94


if __name__ == "__main__":
    c = gf.components.ring()
    co = offset(c, distance=0.5)
    co.show()



"""Returns Component from OASIS file.

based on https://github.com/atait/phidl.git

"""

import os
from typing import Optional

from gdsfactory.read.import_gds import import_gds
from gdsfactory.types import Component, PathType


def import_oas(
    filename: PathType, cellname: Optional[str] = None, flatten: bool = False
) -> Component:
    """Import Component from OASIS.

    Args:
        filename: filepath for oasis file.
        cellname: to import. Defaults to TOP cell.
        flatten: flattens hierarchy.

    """
    filename = str(filename)

    if filename.lower().endswith(".gds"):
        # you are looking for import_gds
        retval = import_gds(filename, cellname=cellname, flatten=flatten)
        return retval
    try:
        import klayout.db as pya
    except ImportError as err:
        err.args = (
            "you need klayout package to read OASIS"
            "pip install klayout\n" + err.args[0],
        ) + err.args[1:]
        raise
    if not filename.lower().endswith(".oas"):
        filename += ".oas"
    fileroot = os.path.splitext(filename)[0]
    tempfilename = f"{fileroot}-tmp.gds"

    layout = pya.Layout()
    layout.read(filename)

    # We want to end up with one Component. If the imported layout has multiple top cells,
    # a new toplevel is created, and they go into the second level
    if len(layout.top_cells()) > 1:
        topcell = layout.create_cell("toplevel")
        rot_DTrans = pya.DTrans.R0
        origin = pya.DPoint(0, 0)
        for childcell in layout.top_cells():
            if childcell == topcell:
                continue
            topcell.insert(
                pya.DCellInstArray(
                    childcell.cell_index(), pya.DTrans(rot_DTrans, origin)
                )
            )
    else:
        topcell = layout.top_cell()
    topcell.write(tempfilename)
    retval = import_gds(tempfilename, cellname=cellname, flatten=flatten)
    os.remove(tempfilename)
    return retval


if __name__ == "__main__":
    c = import_oas(filename="/home/jmatres/demo/a.oas")
    c.show()



import gdsfactory as gf
from gdsfactory.component import Component


def test_netlist_read() -> Component:
    filepath = gf.CONFIG["netlists"] / "mzi.yml"
    c = gf.read.from_yaml(filepath)

    # print(c.get_netlist().pretty())
    # print((c.get_netlist().connections.pretty()))
    # print(len(c.get_netlist().connections))
    # print(len(c.get_dependencies()))
    # assert len(c.get_netlist().connections) == 18

    assert len(c.get_dependencies()) == 5, len(c.get_dependencies())
    return c


def regenerate_regression_test() -> None:
    c = gf.components.mzi()

    filepath = gf.CONFIG["netlists"] / "mzi.yml"
    c.write_netlist(filepath)


if __name__ == "__main__":
    c = test_netlist_read()
    # test_netlist_read_full()
    # regenerate_regression_test()

    # c = test_netlist_read_full()
    c.show(show_ports=True)

    # n = c.get_netlist()
    # i = n["instances"]
    # b = i["bend_circular_R10_17.873_-5.5"]
    # layer = b["settings"]["layer"]
    # print(type(layer))
    # c.show(show_ports=True)



from functools import partial
from typing import Optional

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.port import select_ports
from gdsfactory.routing.get_route import get_route
from gdsfactory.types import ComponentSpec, CrossSectionSpec

default_mzm = dict(
    component="mzi",
    settings=dict(
        straight_x_top="straight_pin",
        cross_section_x_top="pin",
        delta_length=10.0,
    ),
)


@cell
def coh_tx_single_pol(
    balanced_phase_shifters: bool = False,
    mzm_y_spacing: float = 50.0,
    phase_shifter: ComponentSpec = "straight_pin",
    phase_shifter_length: float = 100.0,
    mzm_ps_spacing: float = 40.0,
    splitter: ComponentSpec = "mmi1x2",
    combiner: Optional[ComponentSpec] = None,
    mzm: ComponentSpec = default_mzm,
    mzm_length: float = 200.0,
    with_pads: bool = True,
    xspacing: float = 40.0,
    input_coupler: Optional[ComponentSpec] = None,
    output_coupler: Optional[ComponentSpec] = None,
    pad_array: ComponentSpec = "pad_array",
    cross_section: CrossSectionSpec = "strip",
    **kwargs
) -> Component:
    """MZM-based single polarization coherent transmitter.

    Args:
        balanced_phase_shifters: True adds phase sifters after the MZM at both the I and Q arms.
            False, only adds Q arm has a phase shifter.
        mzm_y_spacing: vertical spacing between the bottom of the I MZM and the top of the Q MZM.
        phase_shifter: phase_shifter spec.
        phase_shifter_length: length of the phase shifter.
        mzm_ps_spacing: spacing between the end of the mzm and the phase shifter.
        splitter: splitter spec.
        combiner: combiner spec.
        mzm: Mach-Zehnder modulator spec.
        mzm_length: length of the MZMs.
        with_pads: if True, we draw pads for all the electrical contacts.
        xspacing: horizontal spacing between the splitter and combiner and the mzm.
        input_coupler: Optional coupler to add before the splitter.
        output_coupler: Optioncal coupler to add after the combiner.
        pad_array: array of pads spec.
        cross_section: for routing (splitter to mzms and mzms to combiners).
        kwargs: cross_section settings.

    .. code::

                                ___ mzm_i __ ps_i__
                                |                  |
                                |                  |
                                |                  |
       (in_coupler)---splitter==|                  |==combiner---(out_coupler)
                                |                  |
                                |                  |
                                |___ mzm_q __ ps_q_|
    """
    from gdsfactory.components.straight import straight as straight_function

    combiner = combiner or splitter
    pad_array = dict(component=pad_array, settings=dict(columns=1, rows=1))

    # ----- Draw MZIs -----
    c = Component()

    if with_pads:
        mzm_mod = gf.routing.add_electrical_pads_top(
            component=gf.get_component(mzm, length_x=mzm_length),
            direction="right",
            spacing=(-100, -136),
            select_ports=partial(select_ports, names=["e1", "e2"]),
            pad_array=pad_array,
        )
        mzm_mod_p = gf.routing.add_electrical_pads_top(
            component=mzm_mod,
            direction="right",
            spacing=(100, -250),
            select_ports=partial(select_ports, names=["e3", "e4"], clockwise=False),
            pad_array=pad_array,
        )
    else:
        mzm_mod_p = mzm(length_x=mzm_length)

    mzm_i = c << mzm_mod_p
    mzm_q = c << mzm_mod_p

    # Separate the two mzms so they don't overlap
    mzm_q.movey(mzm_i.ymin - mzm_y_spacing - mzm_q.ymax)

    # ------------ Phase shifter addition (for I/Q arm definition) ----------------

    phase_shifter = gf.get_component(phase_shifter, length=phase_shifter_length)

    if with_pads:
        ps_w_pads = gf.routing.add_electrical_pads_top(
            component=phase_shifter,
            direction="right",
            spacing=(50, -136),
            select_ports=partial(select_ports, names=["bot_e2", "top_e2"]),
            pad_array=pad_array,
        )
    else:
        ps_w_pads = phase_shifter

    if balanced_phase_shifters:
        ps_i = c << ps_w_pads
    else:
        # only the q arm has a phase shifter
        straight = straight_function(
            length=phase_shifter_length, cross_section=cross_section, **kwargs
        )
        ps_i = c << straight
    ps_q = c << ps_w_pads

    # Connect to the right ports
    if mzm_ps_spacing <= 0.0:
        ps_i.connect("o1", mzm_i.ports["o2"])
        ps_q.connect("o1", mzm_q.ports["o2"])
    else:
        straight_conn = straight_function(
            length=mzm_ps_spacing, cross_section=cross_section, **kwargs
        )
        straight_i = c << straight_conn
        straight_q = c << straight_conn
        straight_i.connect("o1", mzm_i.ports["o2"])
        straight_q.connect("o1", mzm_q.ports["o2"])
        ps_i.connect("o1", straight_i.ports["o2"])
        ps_q.connect("o1", straight_q.ports["o2"])

    # ------------ Splitters and combiners ---------------

    splitter = gf.get_component(splitter)
    sp = c << splitter
    sp.x = mzm_q.xmin - xspacing
    sp.y = (mzm_i.ports["o1"].y + mzm_q.ports["o1"].y) / 2

    route = get_route(
        sp.ports["o2"],
        mzm_i.ports["o1"],
        cross_section=cross_section,
        with_sbend=False,
        **kwargs
    )
    c.add(route.references)

    route = get_route(
        sp.ports["o3"],
        mzm_q.ports["o1"],
        cross_section=cross_section,
        with_sbend=False,
        **kwargs
    )
    c.add(route.references)

    combiner = gf.get_component(combiner)
    comb = c << combiner
    comb.mirror()

    comb.x = ps_q.xmax + xspacing
    comb.y = (mzm_i.ports["o2"].y + mzm_q.ports["o2"].y) / 2

    route = get_route(
        comb.ports["o2"],
        ps_i.ports["o2"],
        cross_section=cross_section,
        with_sbend=False,
        **kwargs
    )
    c.add(route.references)

    route = get_route(
        comb.ports["o3"],
        ps_q.ports["o2"],
        cross_section=cross_section,
        with_sbend=False,
        **kwargs
    )
    c.add(route.references)

    # ------- In and out couplers (if indicated) -----

    if input_coupler is not None:
        # Add input coupler
        in_coupler = gf.get_component(input_coupler)
        in_coup = c << in_coupler
        in_coup.connect("o1", sp.ports["o1"])

    else:
        c.add_port("o1", port=sp.ports["o1"])

    if output_coupler is not None:
        # Add output coupler
        output_coupler = gf.get_component(output_coupler)
        out_coup = c << output_coupler
        out_coup.connect("o1", comb.ports["o1"])
    else:
        c.add_port("o2", port=comb.ports["o1"])

    # ------ Extract electrical ports (if no pads) -------
    if not with_pads:
        c.add_ports(ps_i.get_ports_list(port_type="electrical"), prefix="ps_i")
        c.add_ports(ps_q.get_ports_list(port_type="electrical"), prefix="ps_q")
        c.add_ports(mzm_i.get_ports_list(port_type="electrical"), prefix="mzm_i")
        c.add_ports(mzm_q.get_ports_list(port_type="electrical"), prefix="mzm_q")
        c.auto_rename_ports()

    return c


if __name__ == "__main__":
    c = coh_tx_single_pol(with_pads=True)
    c.show(show_ports=True)



from gdsfactory.types import ComponentModel, Dict, List, Optional


class Schematic:
    instances: Optional[Dict[str, ComponentModel]] = None
    nets: Optional[List[List[str]]] = None
    ports: Optional[Dict[str, str]] = None


if __name__ == "__main__":
    s = Schematic()



"""ensures gdslibrary exists."""

import gdsfactory as gf


def test_gdslib_exists() -> None:
    assert gf.CONFIG["gdslib"].exists()


if __name__ == "__main__":
    test_gdslib_exists()
    # print(gf.CONFIG['gdslib'])



import jsondiff
import numpy as np
import pytest
from omegaconf import OmegaConf
from pytest_regressions.data_regression import DataRegressionFixture

from gdsfactory.component import Component
from gdsfactory.difftest import difftest
from gdsfactory.read.from_yaml import from_yaml, sample_doe_function, sample_mmis

sample_connections = """
name: sample_connections

instances:
    wgw:
      component: straight
      settings:
        length: 1
    wgn:
      component: straight
      settings:
        length: 0.5

connections:
    wgw,o1: wgn,o2

"""

#
#        __Lx__
#       |      |
#       Ly     Lyr
#       |      |
#  CP1==|      |==CP2
#       |      |
#       Ly     Lyr
#       |      |
#      DL/2   DL/2
#       |      |
#       |__Lx__|
#


sample_mirror_simple = """
name: sample_mirror_simple

instances:
    s:
        component: straight

    b:
        component: bend_circular

placements:
    b:
        mirror: True
        port: o1

connections:
    b,o1: s,o2

"""


def test_sample() -> Component:
    c = from_yaml(sample_mmis)
    assert len(c.get_dependencies()) == 6, len(c.get_dependencies())
    assert len(c.ports) == 3, len(c.ports)
    return c


def test_connections() -> Component:
    c = from_yaml(sample_connections)
    assert len(c.get_dependencies()) == 2
    assert len(c.ports) == 0
    return c


sample_2x2_connections = """
name: connections_2x2_solution

instances:
    mmi_bottom:
      component: mmi2x2
      settings:
            length_mmi: 5
    mmi_top:
      component: mmi2x2
      settings:
            length_mmi: 10

placements:
    mmi_top:
        x: 100
        y: 100

routes:
    optical:
        links:
            mmi_bottom,o4: mmi_top,o1
            mmi_bottom,o3: mmi_top,o2

        settings:
            cross_section:
                cross_section: strip
                settings:
                    layer: [2, 0]

"""


def test_connections_2x2() -> Component:
    c = from_yaml(sample_2x2_connections)
    assert len(c.get_dependencies()) == 8, len(c.get_dependencies())
    assert len(c.ports) == 0, len(c.ports)

    length = c.routes["mmi_bottom,o3:mmi_top,o2"]
    assert np.isclose(length, 168.274), length
    return c


sample_different_factory = """
name: sample_different_factory

instances:
    bl:
      component: pad
    tl:
      component: pad
    br:
      component: pad
    tr:
      component: pad

placements:
    tl:
        x: 0
        y: 200

    br:
        x: 400
        y: 400

    tr:
        x: 400
        y: 600

routes:
    electrical:
        settings:
            separation: 20
            cross_section:
                cross_section: metal3_with_bend
                settings:
                    layer: [31, 0]
                    width: 10
        links:
            tl,e3: tr,e1
            bl,e3: br,e1
    optical:
        settings:
            cross_section:
                cross_section: strip
                settings:
                    radius: 100
        links:
            bl,e4: br,e3

"""


def test_connections_different_factory() -> Component:
    c = from_yaml(sample_different_factory)
    lengths = [693.274, 693.274, 1199.144]

    assert np.isclose(c.routes["tl,e3:tr,e1"], lengths[0]), c.routes["tl,e3:tr,e1"]
    assert np.isclose(c.routes["bl,e3:br,e1"], lengths[1]), c.routes["bl,e3:br,e1"]
    assert np.isclose(c.routes["bl,e4:br,e3"], lengths[2]), c.routes["bl,e4:br,e3"]

    return c


sample_different_link_factory = """
name: sample_path_length_matching

instances:
    bl:
      component: pad
    tl:
      component: pad
    br:
      component: pad
    tr:
      component: pad

placements:
    tl:
        x: 0
        y: 200

    br:
        x: 900
        y: 400

    tr:
        x: 900
        y: 600

routes:
    route1:
        routing_strategy: get_bundle_path_length_match
        settings:
            radius: 10
            extra_length: 500
        links:
            tl,e3: tr,e1
            bl,e3: br,e1

"""


def test_connections_different_link_factory() -> Component:
    c = from_yaml(sample_different_link_factory)

    length = 1719.822
    assert np.isclose(c.routes["tl,e3:tr,e1"], length), c.routes["tl,e3:tr,e1"]
    assert np.isclose(c.routes["bl,e3:br,e1"], length), c.routes["bl,e3:br,e1"]
    return c


sample_waypoints = """
name: sample_waypoints

instances:
    t:
      component: pad_array
      settings:
          orientation: 270
    b:
      component: pad_array
      settings:
          orientation: 90

placements:
    t:
        x: -250
        y: 1000
routes:
    route1:
        routing_strategy: get_bundle_from_waypoints
        settings:
            waypoints:
                - [0, 300]
                - [400, 300]
                - [400, 400]
                - [-250, 400]
            auto_widen: False
        links:
            b,e11: t,e11
            b,e12: t,e12
"""


sample_docstring = """
name: sample_docstring

instances:
    mmi_bot:
      component: mmi1x2
      settings:
        width_mmi: 5
        length_mmi: 11
    mmi_top:
      component: mmi1x2
      settings:
        width_mmi: 6
        length_mmi: 22

placements:
    mmi_top:
        port: o1
        x: 0
        y: 0
    mmi_bot:
        port: o1
        x: mmi_top,o2
        y: mmi_top,o2
        dx: 40
        dy: -40
routes:
    optical:
        links:
            mmi_top,o3: mmi_bot,o1
"""


sample_regex_connections = """
name: sample_regex_connections

instances:
    left:
      component: nxn
      settings:
        west: 0
        east: 3
        ysize: 20
    right:
      component: nxn
      settings:
        west: 3
        east: 0
        ysize: 20

placements:
    right:
        x: 20
routes:
    optical:
        links:
            left,o:1:3: right,o:1:3
"""

sample_regex_connections_backwards = """
name: sample_regex_connections_backwards

instances:
    left:
      component: nxn
      settings:
        west: 0
        east: 3
        ysize: 20
    right:
      component: nxn
      settings:
        west: 3
        east: 0
        ysize: 20

placements:
    right:
        x: 20
routes:
    optical:
        links:
            left,o:3:1: right,o:3:1
"""


def test_connections_regex() -> Component:
    c = from_yaml(sample_regex_connections)
    route_names = ["left,o1:right,o1", "left,o2:right,o2", "left,o3:right,o3"]

    length = 12.0
    for route_name in route_names:
        assert np.isclose(c.routes[route_name], length)
    return c


def test_connections_regex_backwargs() -> Component:
    c = from_yaml(sample_regex_connections_backwards)
    route_names = ["left,o1:right,o1", "left,o2:right,o2", "left,o3:right,o3"]

    length = 12.0
    for route_name in route_names:
        assert np.isclose(c.routes[route_name], length), c.routes[route_name]
    return c


def test_connections_waypoints() -> Component:
    c = from_yaml(sample_waypoints)

    length = 2036.548
    route_name = "b,e11:t,e11"
    assert np.isclose(c.routes[route_name], length), c.routes[route_name]
    return c


def test_docstring_sample() -> Component:
    c = from_yaml(sample_docstring)
    route_name = "mmi_top,o3:mmi_bot,o1"
    length = 72.024
    assert np.isclose(c.routes[route_name], length), c.routes[route_name]
    return c


yaml_fail = """
name: yaml_fail
instances:
    mmi_long:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 10
    mmi_short:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 5

placements:
    mmi_short:
        port: o1
        x: mmi_long,o2
        y: mmi_long,o2
    mmi_long:
        port: o1
        x: mmi_short,o2
        y: mmi_short,o2
        dx : 10
        dy: 20
"""
#                      ______
#                     |      |
#           dx   1----| short|
#                |    |______|
#                | dy
#   ______ north |
#   |     |
#   |long |
#   |_____|
#       east

yaml_anchor = """
name: yaml_anchor
instances:
    mmi_long:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 10
    mmi_short:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 5

placements:
    mmi_short:
        port: o3
        x: 0
        y: 0
    mmi_long:
        port: o1
        x: mmi_short,east
        y: mmi_short,north
        dx : 10
        dy: 10
"""

sample_doe = """
name: mask

instances:
    mmi1x2_sweep:
       component: pack_doe
       settings:
         doe: mmi1x2
         do_permutations: True
         spacing: 100
         settings:
           length_mmi: [2, 100]
           width_mmi: [4, 10]
"""

sample_doe_grid = """
name: mask_grid

instances:
    mmi1x2_sweep:
       component: pack_doe_grid
       settings:
         doe: mmi1x2
         do_permutations: True
         spacing: [100, 100]
         shape: [2, 2]
         settings:
           length_mmi: [2, 100]
           width_mmi: [4, 10]
"""

sample_rotation = """
name: sample_rotation

instances:
  r1:
    component: rectangle
    settings:
        size: [4, 2]
  r2:
    component: rectangle
    settings:
        size: [2, 4]

placements:
    r1:
        xmin: 0
        ymin: 0
    r2:
        rotation: -90
        xmin: r1,east
        ymin: 0

"""

# FIXME: Fix both unconmmented cases
# yaml_fail should actually fail
# sample_different_factory: returns a zero length straight that gives an error
# when extracting the netlist

yaml_strings = dict(
    yaml_anchor=yaml_anchor,
    # yaml_fail=yaml_fail,
    # sample_regex_connections_backwards=sample_regex_connections_backwards,
    # sample_regex_connections=sample_regex_connections,
    sample_docstring=sample_docstring,
    sample_waypoints=sample_waypoints,
    sample_different_link_factory=sample_different_link_factory,
    # sample_different_factory=sample_different_factory,
    sample_mirror_simple=sample_mirror_simple,
    sample_connections=sample_connections,
    sample_mmis=sample_mmis,
    sample_doe=sample_doe,
    # sample_doe_grid=sample_doe_grid,
    sample_doe_function=sample_doe_function,
    sample_rotation=sample_rotation,
)


@pytest.mark.parametrize("yaml_key", yaml_strings.keys())
def test_gds(yaml_key: str, data_regression: DataRegressionFixture) -> None:
    """Avoid regressions in GDS geometry shapes and layers."""
    yaml_string = yaml_strings[yaml_key]
    c = from_yaml(yaml_string)
    difftest(c)


@pytest.mark.parametrize("yaml_key", yaml_strings.keys())
def test_settings(
    yaml_key: str, data_regression: DataRegressionFixture, check: bool = True
) -> Component:
    """Avoid regressions when exporting settings."""
    yaml_string = yaml_strings[yaml_key]
    c = from_yaml(yaml_string)

    if check:
        data_regression.check(c.to_dict())
    return c


@pytest.mark.parametrize("yaml_key", yaml_strings.keys())
def test_netlists(
    yaml_key: str,
    data_regression: DataRegressionFixture,
    check: bool = True,
) -> None:
    """Write netlists for hierarchical circuits. Checks that both netlists are
    the same jsondiff does a hierarchical diff Component -> netlist ->
    Component -> netlist.

    Args:
        yaml_key: to test.
        data_regression: for regression test.
        check: False, skips test.

    """
    yaml_string = yaml_strings[yaml_key]
    c = from_yaml(yaml_string)
    n = c.get_netlist()
    if check:
        data_regression.check(n)

    yaml_str = OmegaConf.to_yaml(n, sort_keys=True)

    # print(yaml_str)
    c2 = from_yaml(yaml_str, name=c.name)
    n2 = c2.get_netlist()
    # pprint(d)
    # assert len(d) == 0, pprint(d)
    return jsondiff.diff(n, n2)


def _demo_netlist() -> None:
    """path on the route."""
    import gdsfactory as gf

    # c = from_yaml(sample_2x2_connections)
    c = from_yaml(sample_waypoints)
    c = from_yaml(sample_different_factory)
    c.show(show_ports=True)
    n = c.get_netlist()
    yaml_str = OmegaConf.to_yaml(n, sort_keys=True)
    c2 = from_yaml(yaml_str)
    n2 = c2.get_netlist()
    d = jsondiff.diff(n, n2)
    assert len(d) == 0
    gf.show(c2)


if __name__ == "__main__":
    c = test_connections_different_factory()
    c.show()

    # c = test_netlists("sample_mmis", None, False)
    # yaml_key = "sample_doe_function"
    # yaml_key = "sample_mmis"
    # yaml_string = yaml_strings[yaml_key]
    # c = from_yaml(yaml_string)
    # n = c.get_netlist()
    # yaml_str = OmegaConf.to_yaml(n, sort_keys=True)
    # c2 = from_yaml(yaml_str)
    # n2 = c2.get_netlist()
    # d = jsondiff.diff(n, n2)
    # pprint(d)
    # c2.show()



import warnings

from gdsfactory.component import Component
from gdsfactory.config import __version__


def grating_coupler(gc: Component) -> None:
    if not gc.info.get("polarization"):
        raise ValueError(f"{gc.name} does not have polarization")

    if gc.info.get("polarization") not in ["te", "tm", "dual"]:
        raise ValueError(f"{gc.name} polarization not 'te' or 'tm' or 'dual")

    if not gc.info.get("wavelength"):
        raise ValueError(f"{gc.name} wavelength does not have wavelength")
    if not (0.5 < gc.info["wavelength"] < 5.0):
        raise ValueError(f"{gc.name} wavelength {gc.wavelength} should be in um")

    if "o1" not in gc.ports:
        warnings.warn(
            f"grating_coupler {gc.name} should have a o1 port. It has {gc.ports}"
        )
    if "o1" in gc.ports and gc.ports["o1"].orientation != 180:
        warnings.warn(
            f"grating_coupler {gc.name} orientation = {gc.ports['o1'].orientation}"
            " should be 180 degrees"
        )


def version(
    requirement: str, current: str = __version__, package_name="gdsfactory"
) -> None:
    """Raises error if current version does not match requirement."""
    import semantic_version

    s = semantic_version.SimpleSpec(requirement)
    if not s.match(semantic_version.Version(current)):
        raise ValueError(
            f"{package_name} requirement {requirement}\n"
            f"not compatible your current installed version {current}\n"
            "you can run:\n"
            f"pip install {package_name} {requirement}\n"
        )


if __name__ == "__main__":
    # version(">=3.8.10")
    version("<=3.8.7")



"""Calculate the effective refractive index for a 1D mode."""

from typing import List

import numpy as np
from scipy.optimize import fsolve
from typing_extensions import Literal


def get_effective_indices(
    ncore: float,
    nsubstrate: float,
    ncladding: float,
    thickness: float,
    wavelength: float,
    polarization: Literal["te", "tm"],
) -> List[float]:
    """Returns the effective refractive indices for a 1D mode.

    Args:
        epsilon_core: Relative permittivity of the film.
        epsilon_substrate: Relative permittivity of the substrate.
        epsilon_cladding: Relative permittivity of the cladding.
        thickness: Thickness of the film in um.
        wavelength: Wavelength in um.
        polarization: Either "te" or "tm".

    .. code::

        -----------------      |
        ncladding             inf
        -----------------      |
        ncore              thickness
        -----------------      |
        nsubstrate            inf
        -----------------      |

    .. code::

        import gdsfactory.simulation as sim

        neffs = sim.get_effective_indices(
            ncore=3.4777,
            ncladding=1.444,
            nsubstrate=1.444,
            thickness=0.22,
            wavelength=1.55,
            polarization="te",
        )

    """
    epsilon_core = ncore**2
    epsilon_cladding = ncladding**2
    epsilon_substrate = nsubstrate**2

    thickness *= 1e-6
    wavelength *= 1e-6

    if polarization == "te":
        tm = False
    elif polarization == "tm":
        tm = True
    else:
        raise ValueError('Polarization must be "te" or "tm"')

    k_0 = 2 * np.pi / wavelength

    def k_f(e_eff):
        return k_0 * np.sqrt(epsilon_core - e_eff) / (epsilon_core if tm else 1)

    def k_s(e_eff):
        return (
            k_0 * np.sqrt(e_eff - epsilon_substrate) / (epsilon_substrate if tm else 1)
        )

    def k_c(e_eff):
        return k_0 * np.sqrt(e_eff - epsilon_cladding) / (epsilon_cladding if tm else 1)

    def objective(e_eff):
        return 1 / np.tan(k_f(e_eff) * thickness) - (
            k_f(e_eff) ** 2 - k_s(e_eff) * k_c(e_eff)
        ) / (k_f(e_eff) * (k_s(e_eff) + k_c(e_eff)))

    # scan roughly for indices
    # use a by 1e-10 smaller search area to avoid division by zero
    x = np.linspace(
        min(epsilon_substrate, epsilon_cladding) + 1e-10, epsilon_core - 1e-10, 1000
    )
    indices_temp = x[np.abs(objective(x)) < 0.1]
    if not len(indices_temp):
        return []

    # and then use fsolve to get exact indices
    indices_temp = fsolve(objective, indices_temp)

    indices = []
    for index in indices_temp:
        if not any(np.isclose(index, i, atol=1e-5) for i in indices):
            indices.append(index)

    return np.sqrt(indices).tolist()


def test_effective_index():
    neff = get_effective_indices(
        ncore=3.4777,
        ncladding=1.444,
        nsubstrate=1.444,
        thickness=0.22,
        wavelength=1.55,
        polarization="te",
    )
    assert np.isclose(neff[0], 2.8494636999424405)


if __name__ == "__main__":
    print(
        get_effective_indices(
            ncore=3.4777,
            ncladding=1.444,
            nsubstrate=1.444,
            thickness=0.22,
            wavelength=1.55,
            polarization="te",
        )
    )



from functools import partial
from typing import Dict, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np

import gdsfactory as gf


def plot_sparameters(
    sp: Dict[str, np.ndarray],
    logscale: bool = True,
    keys: Optional[Tuple[str, ...]] = None,
    with_sparameter_labels: bool = True,
    with_sparameter_keys: bool = False,
) -> None:
    """Plots Sparameters from a dict of np.ndarrays.

    Args:
        sp: Sparameters np.ndarray.
        logscale: plots 20*log10(S).
        keys: list of keys to plot, plots all by default.
        with_sparameter_keys: uses S11, S12 in plot labels.
        with_sparameter_keys: use Sparameter keys.
            Assumes mode 0 and o1, o2, o3 port naming.

    .. plot::
        :include-source:

        import gdsfactory as gf
        import gdsfactory.simulation as sim

        sp = sim.get_sparameters_data_lumerical(component=gf.components.mmi1x2)
        sim.plot.plot_sparameters(sp, logscale=True)

    """
    w = sp["wavelengths"] * 1e3
    keys = keys or [key for key in sp if not key.lower().startswith("wav")]

    for key in keys:

        if with_sparameter_keys:
            key = f"o{key[1]}@0,o{key[2]}@0"

        if with_sparameter_labels and "o" in key and "@" in key:
            port_mode1, port_mode2 = key.split(",")
            port1, _mode1 = port_mode1.split("@")
            port2, _mode2 = port_mode2.split("@")
            alias = f"S{port1[1:]}{port2[1:]}"
        else:
            alias = key

        if key not in sp:
            raise ValueError(f"{key} not in {sp.keys()}")
        y = sp[key]
        y = 20 * np.log10(np.abs(y)) if logscale else np.abs(y) ** 2
        plt.plot(w, y, label=alias)
    plt.legend()
    plt.xlabel("wavelength (nm)")
    plt.ylabel("|S| (dB)") if logscale else plt.ylabel("$|S|^2$")
    plt.show()


def plot_imbalance2x2(
    sp: Dict[str, np.ndarray], port1: str = "o1@0,o3@0", port2: str = "o1@0,o4@0"
) -> None:
    """Plots imbalance in % for 2x2 coupler.

    Args:
        sp: sparameters dict np.ndarray.
        port1: name.
        port2: name.

    """
    y1 = np.abs(sp[port1])
    y2 = np.abs(sp[port2])
    imbalance = y1 / y2
    x = sp["wavelengths"] * 1e3
    plt.plot(x, 100 * abs(imbalance))
    plt.xlabel("wavelength (nm)")
    plt.ylabel("imbalance (%)")
    plt.grid()


def plot_loss2x2(
    sp: Dict[str, np.ndarray], port1: str = "o1@0,", port2: str = "o1@0"
) -> None:
    """Plots imbalance in % for 2x2 coupler.

    Args:
        sp: sparameters dict np.ndarray.
        port1: port name @ mode index. o1@0 is the fundamental mode for o1 port.
        port2: port name @ mode index. o1@0 is the fundamental mode for o1 port.

    """
    y1 = np.abs(sp[port1])
    y2 = np.abs(sp[port2])
    x = sp["wavelengths"] * 1e3
    plt.plot(x, abs(10 * np.log10(y1**2 + y2**2)))
    plt.xlabel("wavelength (nm)")
    plt.ylabel("excess loss (dB)")


plot_loss1x2 = partial(plot_loss2x2, port1="o1@0,o2@0", port2="o1@0,o3@0")
plot_imbalance1x2 = partial(plot_imbalance2x2, port1="o1@0,o2@0", port2="o1@0,o3@0")


if __name__ == "__main__":
    import gdsfactory.simulation as sim

    sp = sim.get_sparameters_data_tidy3d(component=gf.components.mmi1x2)
    plot_sparameters(sp, logscale=False, keys=["o1@0,o2@0"])
    plt.show()



import pytest
from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.difftest import difftest

sequences = ["ABHBA", "!HH", "AB", "HH!", "H"]


@pytest.mark.parametrize("index", range(len(sequences)))
def test_component_from_sequence(
    index: int, data_regression: DataRegressionFixture, check: bool = True
) -> gf.Component:

    bend180 = gf.components.bend_circular180()
    wg_pin = gf.components.straight_pin(length=40)
    wg = gf.components.straight()

    # Define a map between symbols and (component, input port, output port)
    symbol_to_component_map = {
        "A": (bend180, "o1", "o2"),
        "B": (bend180, "o2", "o1"),
        "H": (wg_pin, "o1", "o2"),
        "-": (wg, "o1", "o2"),
    }

    sequence = sequences[index]
    c = gf.components.component_sequence(
        sequence=sequence, symbol_to_component=symbol_to_component_map
    )

    difftest(c)
    if check:
        data_regression.check(c.to_dict())
    return c


if __name__ == "__main__":
    c = test_component_from_sequence(1, None, False)
    c.show()



"""Returns tidy3d simulation from gdsfactory Component."""
import warnings
from typing import Dict, Optional

import matplotlib.pyplot as plt
import numpy as np
import tidy3d as td

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.extension import move_polar_rad_copy
from gdsfactory.config import logger
from gdsfactory.pdk import get_layer_stack
from gdsfactory.simulation.gtidy3d.materials import (
    MATERIAL_NAME_TO_TIDY3D_INDEX,
    MATERIAL_NAME_TO_TIDY3D_NAME,
    get_index,
    get_medium,
)
from gdsfactory.tech import LayerStack


def get_simulation_grating_coupler(
    component: Component,
    port_extension: Optional[float] = 15.0,
    layer_stack: Optional[LayerStack] = None,
    thickness_pml: float = 1.0,
    xmargin: float = 0,
    ymargin: float = 0,
    xmargin_left: float = 0,
    xmargin_right: float = 0,
    ymargin_top: float = 0,
    ymargin_bot: float = 0,
    zmargin: float = 1.0,
    clad_material: str = "sio2",
    box_material: str = "sio2",
    box_thickness: float = 2.0,
    substrate_material: str = "si",
    port_waveguide_name: str = "o1",
    port_margin: float = 0.5,
    port_waveguide_offset: float = 0.1,
    distance_source_to_monitors: float = 0.2,
    wavelength: Optional[float] = 1.55,
    wavelength_start: float = 1.20,
    wavelength_stop: float = 1.80,
    wavelength_points: int = 256,
    plot_modes: bool = False,
    num_modes: int = 2,
    run_time_ps: float = 10.0,
    fiber_port_name: str = "vertical_te",
    fiber_xoffset: float = 0,
    fiber_z: float = 2,
    fiber_mfd: float = 5.2,
    fiber_angle_deg: float = 20.0,
    dispersive: bool = False,
    material_name_to_tidy3d_index: Dict[str, float] = MATERIAL_NAME_TO_TIDY3D_INDEX,
    material_name_to_tidy3d_name: Dict[str, str] = MATERIAL_NAME_TO_TIDY3D_NAME,
    is_3d: bool = True,
    with_all_monitors: bool = False,
    boundary_spec: Optional[td.BoundarySpec] = None,
    grid_spec: Optional[td.GridSpec] = None,
    sidewall_angle_deg: float = 0,
    dilation: float = 0.0,
    **kwargs,
) -> td.Simulation:
    r"""Returns Simulation object from a gdsfactory grating coupler component.

    injects a Gaussian beam from above and monitors the transmission into the waveguide.

    based on grating coupler example
    https://docs.simulation.cloud/projects/tidy3d/en/latest/notebooks/GratingCoupler.html

    .. code::

         top view
              ________________________________
             |                               |
             | xmargin_left                  |
             |<------>                       |
             |           ________________    |
             |          /   |  |  |  |  |    |
             |         /    |  |  |  |  |    |
             |=========     |  |  |  |  |    |
             |         \    |  |  |  |  |    |
             |   _ _ _ _\___|__|__|__|__|    |
             |   |                       <-->|
             |   |ymargin_bot   xmargin_right|
             |   |                           |
             |___|___________________________|


        side view

              fiber_xoffset
                 |<--->|
            fiber_port_name
                 |
                         fiber_angle_deg > 0
                        |  /
                        | /
                        |/
                 /              /       |
                /  fiber_mfd   /        |
               /<------------>/    _ _ _| _ _ _ _ _ _ _
                                        |
                   clad_material        | fiber_z
                    _   _   _      _ _ _|_ _ _ _ _ _ _
                   | | | | | |          ||wg_thickness
                  _| |_| |_| |__________||___
                                        || |
        waveguide            |          || | slab_thickness
              ____________________ _ _ _||_|_
                             |          |
                   box_material         |box_thickness
              _______________|____ _ _ _|_ _ _ _ _ _ _
                             |          |
                 substrate_material     |substrate_thickness
             ________________|____ _ _ _|_ _ _ _ _ _ _
                             |
        |--------------------|<-------->
                                xmargin

    Args:
        component: gdsfactory Component.
        port_extension: extend ports beyond the PML.
        layer_stack: contains layer to thickness, zmin and material.
            Defaults to active pdk.layer_stack.
        thickness_pml: PML thickness (um).
        xmargin: left/right distance from component to PML.
        xmargin_left: left distance from component to PML.
        xmargin_right: right distance from component to PML.
        ymargin: left/right distance from component to PML.
        ymargin_top: top distance from component to PML.
        ymargin_bot: bottom distance from component to PML.
        zmargin: thickness for cladding above and below core.
        clad_material: material for cladding.
        box_material:
        substrate_material:
        box_thickness: (um).
        substrate_thickness: (um).
        port_waveguide_name: input port name.
        port_margin: margin on each side of the port.
        distance_source_to_monitors: in (um) source goes before monitors.
        port_waveguide_offset: mode solver workaround.
            positive moves source forward, negative moves source backward.
        wavelength: source center wavelength (um).
            if None takes mean between wavelength_start, wavelength_stop
        wavelength_start: in (um).
        wavelength_stop: in (um).
        wavelength_points: number of wavelengths.
        plot_modes: plot source modes.
        num_modes: number of modes to plot.
        run_time_ps: make sure it's sufficient for the fields to decay.
            defaults to 10ps and automatic shutoff stops earlier if needed.
        fiber_port_name: for the component.
        fiber_xoffset: fiber center xoffset to fiber_port_name.
        fiber_z: fiber zoffset from grating zmax.
        fiber_mfd: fiber mode field diameter (um).
        fiber_angle_deg: fiber_angle in degrees with respect to normal.
            Positive for west facing, Negative for east facing sources.
        dispersive: False uses constant refractive index materials.
            True adds wavelength depending materials.
            Dispersive materials require more computation.
        material_name_to_tidy3d_index: not dispersive materials have a constant index.
        material_name_to_tidy3d_name: dispersive materials have a wavelength
            dependent index. Maps layer_stack names with tidy3d material database names.
        is_3d: if False collapses the Y direction for a 2D simulation.
        with_all_monitors: True includes field monitors which increase results filesize.
        grid_spec: defaults to automatic td.GridSpec.auto(wavelength=wavelength)
            td.GridSpec.uniform(dl=20*nm)
            td.GridSpec(
                grid_x = td.UniformGrid(dl=0.04),
                grid_y = td.AutoGrid(min_steps_per_wvl=20),
                grid_z = td.AutoGrid(min_steps_per_wvl=20),
                wavelength=wavelength,
                override_structures=[refine_box]
            )
        boundary_spec: Specification of boundary conditions along each dimension.
            Defaults to td.BoundarySpec.all_sides(boundary=td.PML())
        dilation: float = 0.0
            Dilation of the polygon in the base by shifting each edge along its
            normal outwards direction by a distance;
            a negative value corresponds to erosion.
        sidewall_angle_deg : float = 0
            Angle of the sidewall.
            ``sidewall_angle=0`` (default) specifies vertical wall,
            while ``0<sidewall_angle_deg<90`` for the base to be larger than the top.

    keyword Args:
        symmetry: Define Symmetries.
            Tuple of integers defining reflection symmetry across a plane
            bisecting the simulation domain normal to the x-, y-, and z-axis
            at the simulation center of each axis, respectvely.
            Each element can be ``0`` (no symmetry), ``1`` (even, i.e. 'PMC' symmetry) or
            ``-1`` (odd, i.e. 'PEC' symmetry).
            Note that the vectorial nature of the fields must be taken into account to correctly
            determine the symmetry value.
        medium: Background medium of simulation, defaults to vacuum if not specified.
        shutoff: shutoff condition
            Ratio of the instantaneous integrated E-field intensity to the maximum value
            at which the simulation will automatically terminate time stepping.
            Used to prevent extraneous run time of simulations with fully decayed fields.
            Set to ``0`` to disable this feature.
        subpixel: subpixel averaging.If ``True``, uses subpixel averaging of the permittivity
        based on structure definition, resulting in much higher accuracy for a given grid size.
        courant: courant factor.
            Courant stability factor, controls time step to spatial step ratio.
            Lower values lead to more stable simulations for dispersive materials,
            but result in longer simulation times.
        version: String specifying the front end version number.

    .. code::

        import matplotlib.pyplot as plt
        import gdsfactory as gf
        import gdsfactory.simulation.gtidy3d as gt

        c = gf.components.grating_coupler_elliptical_arbitrary(
            widths=[0.343] * 25, gaps=[0.345] * 25
        )
        sim = gt.get_simulation(c)
        gt.plot_simulation(sim)

    """
    layer_stack = layer_stack or get_layer_stack()

    layer_to_thickness = layer_stack.get_layer_to_thickness()
    layer_to_material = layer_stack.get_layer_to_material()
    layer_to_zmin = layer_stack.get_layer_to_zmin()
    # layer_to_sidewall_angle = layer_stack.get_layer_to_sidewall_angle()

    boundary_spec = boundary_spec or td.BoundarySpec.all_sides(boundary=td.PML())
    grid_spec = grid_spec or td.GridSpec.auto(wavelength=wavelength)

    if dispersive:
        material_name_to_tidy3d = material_name_to_tidy3d_name
    else:
        material_name_to_tidy3d = material_name_to_tidy3d_index

    assert isinstance(
        component, Component
    ), f"component needs to be a gf.Component, got Type {type(component)}"

    if port_waveguide_name not in component.ports:
        warnings.warn(
            f"port_waveguide_name={port_waveguide_name} not in {component.ports.keys()}"
        )
        port_waveguide = component.get_ports_list()[0]
        port_waveguide_name = port_waveguide.name
        warnings.warn(f"Selecting port_waveguide_name={port_waveguide_name} instead.")

    if fiber_port_name not in component.ports:
        raise ValueError(
            f"fiber_port_name = {fiber_port_name!r} not in {component.ports.keys()}"
        )

    component_padding = gf.add_padding_container(
        component,
        default=0,
        top=ymargin or ymargin_top,
        bottom=ymargin or ymargin_bot,
        left=xmargin or xmargin_left,
        right=xmargin or xmargin_right,
    )
    component_extended = (
        gf.components.extension.extend_ports(
            component=component_padding, length=port_extension, centered=True
        )
        if port_extension
        else component
    )

    gf.show(component_extended)
    component_extended = component_extended.flatten()
    component_ref = component_padding.ref()
    component_ref.x = 0
    component_ref.y = 0

    layers_thickness = [
        layer_to_thickness[layer]
        for layer in component.get_layers()
        if layer in layer_to_thickness
    ]

    if len(layer_to_thickness) < 1:
        raise ValueError(f"{component.get_layers()} not in {layer_to_thickness.keys()}")

    wg_thickness = max(layers_thickness)
    sim_xsize = component_ref.xsize + 2 * thickness_pml
    sim_zsize = (
        thickness_pml + box_thickness + wg_thickness + thickness_pml + 2 * zmargin
    )
    sim_ysize = component_ref.ysize + 2 * thickness_pml if is_3d else 0
    sim_size = [
        sim_xsize,
        sim_ysize,
        sim_zsize,
    ]

    clad_material_name_or_index = material_name_to_tidy3d[clad_material]
    box_material_name_or_index = material_name_to_tidy3d[box_material]
    substrate_material_name_or_index = material_name_to_tidy3d[substrate_material]

    clad = td.Structure(
        geometry=td.Box(
            size=(td.inf, td.inf, sim_zsize),
            center=(0, 0, sim_zsize / 2),
        ),
        medium=get_medium(name_or_index=clad_material_name_or_index),
    )
    box = td.Structure(
        geometry=td.Box(
            size=(td.inf, td.inf, box_thickness),
            center=(0, 0, -box_thickness / 2),
        ),
        medium=get_medium(name_or_index=box_material_name_or_index),
    )

    substrate_thickness = 10
    substrate = td.Structure(
        geometry=td.Box(
            size=(td.inf, td.inf, substrate_thickness),
            center=(0, 0, -box_thickness - substrate_thickness / 2),
        ),
        medium=get_medium(name_or_index=substrate_material_name_or_index),
    )

    structures = [substrate, box, clad]

    for layer in component.layers:
        if layer in layer_to_thickness and layer in layer_to_material:
            thickness = layer_to_thickness[layer]
            zmin = layer_to_zmin[layer]
            zmax = zmin + thickness
            if layer_to_material[layer] in material_name_to_tidy3d:
                name_or_index = material_name_to_tidy3d[layer_to_material[layer]]
                medium = get_medium(name_or_index=name_or_index)
                index = get_index(name_or_index=name_or_index)
                logger.debug(
                    f"Add {layer}, {name_or_index!r}, index = {index:.3f}, "
                    f"thickness = {thickness}, zmin = {zmin}, zmax = {zmax}"
                )

                polygons = td.PolySlab.from_gds(
                    gds_cell=component_extended,
                    gds_layer=layer[0],
                    gds_dtype=layer[1],
                    axis=2,
                    slab_bounds=(zmin, zmax),
                    sidewall_angle=np.deg2rad(sidewall_angle_deg),
                    dilation=dilation,
                )

                for polygon in polygons:
                    geometry = td.Structure(
                        geometry=polygon,
                        medium=medium,
                    )
                    structures.append(geometry)
            elif layer not in layer_to_material:
                logger.debug(f"Layer {layer} not in {layer_to_material.keys()}")
            else:
                materials = list(material_name_to_tidy3d.keys())
                logger.debug(f"material {layer_to_material[layer]} not in {materials}")

    wavelengths = np.linspace(wavelength_start, wavelength_stop, wavelength_points)
    freqs = td.constants.C_0 / wavelengths
    freq0 = td.constants.C_0 / np.mean(wavelengths)
    fwidth = freq0 / 10

    # Add input waveguide port
    port = component_ref.ports[port_waveguide_name]
    angle = port.orientation
    width = port.width + 2 * port_margin
    size_x = width * abs(np.sin(angle * np.pi / 180))
    size_y = width * abs(np.cos(angle * np.pi / 180))
    size_x = 0 if size_x < 0.001 else size_x
    size_y = 0 if size_y < 0.001 else size_y
    size_y = size_y if is_3d else td.inf
    size_z = wg_thickness + 2 * zmargin
    waveguide_port_size = [size_x, size_y, size_z]
    xy_shifted = move_polar_rad_copy(
        np.array(port.center), angle=angle * np.pi / 180, length=port_waveguide_offset
    )
    waveguide_port_center = xy_shifted.tolist() + [0]  # (x, y, z=0)

    waveguide_monitor = td.ModeMonitor(
        center=waveguide_port_center,
        size=waveguide_port_size,
        freqs=freqs,
        mode_spec=td.ModeSpec(num_modes=1),
        name="waveguide",
    )

    # Add fiber monitor
    fiber_port = component_ref.ports[fiber_port_name]
    fiber_port_x = fiber_port.x + fiber_xoffset

    assert -sim_size[0] / 2 < fiber_port_x < sim_size[0] / 2, (
        f"component.ports[{fiber_port_name!r}] + (fiber_xoffset = {fiber_xoffset}). "
        f"{fiber_port_x} needs to be between {-sim_size[0]/2} and {+sim_size[0]/2}"
    )

    # inject Gaussian beam from above and monitors the transmission into the waveguide.
    gaussian_beam = td.GaussianBeam(
        size=(td.inf, td.inf, 0),
        center=[fiber_port_x, 0, fiber_z],
        source_time=td.GaussianPulse(freq0=freq0, fwidth=fwidth),
        angle_theta=np.deg2rad(-fiber_angle_deg),
        angle_phi=np.pi,
        direction="-",
        waist_radius=fiber_mfd / 2,
        pol_angle=np.pi / 2,
    )

    plane_monitor = td.FieldMonitor(
        center=[0, 0, (zmax + zmin) / 2],
        size=[sim_size[0], sim_size[1], 0],
        freqs=[freq0],
        name="full_domain_fields",
    )

    rad_monitor = td.FieldMonitor(
        center=[0, 0, 0],
        size=[td.inf, 0, td.inf],
        freqs=[freq0],
        name="radiated_fields",
    )

    near_field_monitor = td.FieldMonitor(
        center=[0, 0, fiber_z],
        size=[td.inf, td.inf, 0],
        freqs=[freq0],
        name="radiated_near_fields",
    )

    monitors = [waveguide_monitor]
    monitors += (
        [plane_monitor, rad_monitor, near_field_monitor] if with_all_monitors else []
    )

    sim = td.Simulation(
        size=sim_size,
        structures=structures,
        sources=[gaussian_beam],
        monitors=monitors,
        run_time=run_time_ps * 1e-12,
        boundary_spec=boundary_spec,
        **kwargs,
    )

    if plot_modes:
        src_plane = td.Box(center=waveguide_port_center, size=waveguide_port_size)
        ms = td.plugins.ModeSolver(simulation=sim, plane=src_plane, freq=freq0)

        mode_spec = td.ModeSpec(num_modes=num_modes)
        modes = ms.solve(mode_spec=mode_spec)

        print(
            "Effective index of computed modes: ",
            ", ".join([f"{mode.n_eff:1.4f}" for mode in modes]),
        )

        if is_3d:
            fig, axs = plt.subplots(num_modes, 2, figsize=(12, 12))
        else:
            fig, axs = plt.subplots(num_modes, 3, figsize=(12, 12))

        for mode_ind in range(num_modes):
            if is_3d:
                abs(modes[mode_ind].field_data.Ey).plot(
                    x="y", y="z", cmap="magma", ax=axs[mode_ind, 0]
                )
                abs(modes[mode_ind].field_data.Ez).plot(
                    x="y", y="z", cmap="magma", ax=axs[mode_ind, 1]
                )
            else:
                abs(modes[mode_ind].field_data.Ex).plot(ax=axs[mode_ind, 0])
                abs(modes[mode_ind].field_data.Ey).plot(ax=axs[mode_ind, 1])
                abs(modes[mode_ind].field_data.Ez).plot(ax=axs[mode_ind, 2])

                axs[mode_ind, 0].set_title(f"|Ex|: mode_index={mode_ind}")
                axs[mode_ind, 1].set_title(f"|Ey|: mode_index={mode_ind}")
                axs[mode_ind, 2].set_title(f"|Ez|: mode_index={mode_ind}")

        if is_3d:
            axs[mode_ind, 0].set_aspect("equal")
            axs[mode_ind, 1].set_aspect("equal")
        plt.show()
    return sim


if __name__ == "__main__":
    import gdsfactory.simulation.gtidy3d as gt

    c = gf.components.grating_coupler_elliptical_arbitrary(
        widths=[0.343] * 25, gaps=[0.345] * 25
    )
    sim = get_simulation_grating_coupler(
        c,
        plot_modes=False,
        is_3d=False,
        fiber_angle_deg=20,
    )
    gt.plot_simulation(sim)  # make sure simulations looks good

    # c = gf.components.grating_coupler_elliptical_lumerical()  # inverse design grating
    # sim = get_simulation_grating_coupler(c, plot_modes=False, fiber_angle_deg=-5)
    # sim_data = gt.get_results(sim).result()
    # freq0 = td.constants.C_0 / 1.55
    # fig, (ax1, ax2, ax3) = plt.subplots(3, 1, tight_layout=True, figsize=(14, 16))
    # sim_data.plot_field("full_domain_fields", "Ey", freq=freq0, z=0, ax=ax1)
    # sim_data.plot_field("radiated_near_fields", "Ey", freq=freq0, z=0, ax=ax2)
    # sim_data.plot_field("radiated_fields", "Ey", freq=freq0, y=0, ax=ax3)

    # plt.figure()
    # waveguide = sim_data["waveguide"]
    # plt.plot(np.abs(waveguide.amps.values[0].flatten()), label="+")
    # plt.plot(np.abs(waveguide.amps.values[1].flatten()), label="-")
    # plt.legend()
    # plt.show()
    # print(f"waveguide in waveguide / waveguide in = {float(waveguide.amps.values):.2f} ")



"""Wires for electrical manhattan routes."""

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight
from gdsfactory.types import CrossSectionSpec

wire_straight = gf.partial(straight, cross_section="metal3")


@gf.cell
def wire_corner(cross_section: CrossSectionSpec = "metal3", **kwargs) -> Component:
    """Returns 90 degrees electrical corner wire.

    Args:
        cross_section: spec.
        kwargs: cross_section settings.
    """
    x = gf.get_cross_section(cross_section, **kwargs)
    layer = x.layer
    width = x.width

    c = Component()
    a = width / 2
    xpts = [-a, a, a, -a]
    ypts = [-a, -a, a, a]

    c.add_polygon([xpts, ypts], layer=layer)
    c.add_port(
        name="e1",
        center=(-a, 0),
        width=width,
        orientation=180,
        layer=layer,
        port_type="electrical",
    )
    c.add_port(
        name="e2",
        center=(0, a),
        width=width,
        orientation=90,
        layer=layer,
        port_type="electrical",
    )
    c.info["length"] = width
    return c


if __name__ == "__main__":

    # c = wire_straight()
    c = wire_corner()
    # c.show(show_ports=True)
    # c.pprint_ports()
    c.pprint()

    # print(yaml.dump(c.to_dict()))



import gdsfactory as gf
from gdsfactory.add_padding import get_padding_points
from gdsfactory.component import Component
from gdsfactory.path import arc, extrude
from gdsfactory.snap import snap_to_grid
from gdsfactory.types import CrossSectionSpec, LayerSpec


@gf.cell
def bend_circular_heater(
    radius: float = 10,
    angle: float = 90,
    npoints: int = 720,
    heater_to_wg_distance: float = 1.2,
    heater_width: float = 0.5,
    layer_heater: LayerSpec = "HEATER",
    with_bbox: bool = True,
    cross_section: CrossSectionSpec = "strip",
    **kwargs
) -> Component:
    """Creates an arc of arclength `theta` starting at angle `start_angle`.

    Args:
        radius: in um.
        angle: angle of arc (degrees).
        npoints: Number of points used per 360 degrees.
        heater_to_wg_distance: in um.
        heater_width: in um.
        layer_heater: for heater.
        with_bbox: box in bbox_layers and bbox_offsets to avoid DRC sharp edges.
        cross_section: specification (CrossSection, string, CrossSectionFactory dict).
        kwargs: cross_section settings.
    """
    x = gf.get_cross_section(cross_section, radius=radius, **kwargs)
    width = x.width
    layer = x.layer

    offset = heater_to_wg_distance + width / 2
    s1 = gf.Section(
        width=heater_width,
        offset=+offset,
        layer=layer_heater,
    )
    s2 = gf.Section(
        width=heater_width,
        offset=-offset,
        layer=layer_heater,
    )
    x = gf.CrossSection(
        width=width, offset=0, layer=layer, port_names=["in", "out"], sections=[s1, s2]
    )

    p = arc(radius=radius, angle=angle, npoints=npoints)
    c = extrude(p, x)
    c.length = snap_to_grid(p.length())
    c.dx = abs(p.points[0][0] - p.points[-1][0])
    c.dy = abs(p.points[0][0] - p.points[-1][0])

    if with_bbox:
        padding = []
        for offset in x.bbox_offsets:
            top = offset if angle == 180 else 0
            points = get_padding_points(
                component=c,
                default=0,
                bottom=offset,
                right=offset,
                top=top,
            )
            padding.append(points)

        for layer, points in zip(x.bbox_layers, padding):
            c.add_polygon(points, layer=layer)
    return c


if __name__ == "__main__":
    c = bend_circular_heater(heater_width=1)
    print(c.ports)
    c.show(show_ports=True)



"""based on phidl.geometry."""

from typing import Optional, Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.components.text import text
from gdsfactory.types import LayerSpec


@gf.cell
def die(
    size: Tuple[float, float] = (10000.0, 10000.0),
    street_width: float = 100.0,
    street_length: float = 1000.0,
    die_name: Optional[str] = "chip99",
    text_size: float = 100.0,
    text_location: str = "SW",
    layer: LayerSpec = "FLOORPLAN",
    bbox_layer: Optional[LayerSpec] = "FLOORPLAN",
    draw_corners: bool = True,
    draw_dicing_lane: bool = True,
) -> gf.Component:
    """Returns basic die with 4 right angle corners marking the boundary of the.

    chip/die and a label with the name of the die.

    Args:
        size: x, y dimensions of the die.
        street_width: Width of the corner marks for die-sawing.
        street_length: Length of the corner marks for die-sawing.
        die_name: Label text.
        text_size: Label text size.
        text_location: {'NW', 'N', 'NE', 'SW', 'S', 'SE'} Label text compass location.
        layer: Specific layer to put polygon geometry on.
        bbox_layer: optional bbox layer.
        draw_corners: around die.
        draw_dicing_lane: around die.
    """
    c = gf.Component(name="die")
    sx, sy = size[0] / 2, size[1] / 2

    if draw_dicing_lane:
        street_length = max([sx, sy])

    if draw_corners or draw_dicing_lane:
        xpts = np.array(
            [
                sx,
                sx,
                sx - street_width,
                sx - street_width,
                sx - street_length,
                sx - street_length,
            ]
        )
        ypts = np.array(
            [
                sy,
                sy - street_length,
                sy - street_length,
                sy - street_width,
                sy - street_width,
                sy,
            ]
        )
        c.add_polygon([xpts, ypts], layer=layer)
        c.add_polygon([-xpts, ypts], layer=layer)
        c.add_polygon([xpts, -ypts], layer=layer)
        c.add_polygon([-xpts, -ypts], layer=layer)

    if bbox_layer:
        c.add_polygon([[sx, sy], [sx, -sy], [-sx, -sy], [-sx, sy]], layer=bbox_layer)

    if die_name:
        t = c.add_ref(text(text=die_name, size=text_size, layer=layer))

        d = street_width + 20
        if type(text_location) is str:
            text_location = text_location.upper()
            if text_location == "N":
                t.x, t.ymax = [0, sy - d]
            elif text_location == "NE":
                t.xmax, t.ymax = [sx - d, sy - d]
            elif text_location == "NW":
                t.xmin, t.ymax = [-sx + d, sy - d]
            elif text_location == "S":
                t.x, t.ymin = [0, -sy + d]
            elif text_location == "SE":
                t.xmax, t.ymin = [sx - d, -sy + d]
            elif text_location == "SW":
                t.xmin, t.ymin = [-sx + d, -sy + d]
        else:
            t.x, t.y = text_location

    return c


if __name__ == "__main__":
    # c = die(size=(3000, 5000), draw_dicing_lane=True)
    c = die()
    # c.show(show_ports=True)
    c.plot()
    c.show()



from typing import Callable, List, Optional, Tuple, Union

import gdsfactory as gf
from gdsfactory.add_labels import get_input_label_text, get_input_label_text_loopback
from gdsfactory.component import Component, ComponentReference
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.grating_coupler_elliptical_trenches import grating_coupler_te
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.cross_section import strip
from gdsfactory.port import Port, select_ports_optical
from gdsfactory.routing.get_bundle import get_bundle, get_min_spacing
from gdsfactory.routing.get_input_labels import get_input_labels
from gdsfactory.routing.get_route import get_route_from_waypoints
from gdsfactory.routing.manhattan import generate_manhattan_waypoints, round_corners
from gdsfactory.routing.route_south import route_south
from gdsfactory.routing.utils import direction_ports_from_list_ports
from gdsfactory.types import (
    ComponentSpec,
    ComponentSpecOrList,
    CrossSectionSpec,
    Label,
    LayerSpec,
)


def route_fiber_array(
    component: Component,
    fiber_spacing: Union[str, float] = "fiber_array_spacing",
    grating_coupler: ComponentSpecOrList = grating_coupler_te,
    bend: ComponentSpec = bend_euler,
    straight: ComponentSpec = straight_function,
    taper: ComponentSpec = taper_function,
    fanout_length: Optional[float] = None,
    max_y0_optical: None = None,
    with_loopback: bool = True,
    nlabels_loopback: int = 2,
    straight_separation: float = 6.0,
    straight_to_grating_spacing: float = 5.0,
    optical_routing_type: Optional[int] = None,
    connected_port_names: None = None,
    nb_optical_ports_lines: int = 1,
    force_manhattan: bool = False,
    excluded_ports: Optional[List[str]] = None,
    grating_indices: Optional[List[int]] = None,
    route_filter: Callable = get_route_from_waypoints,
    gc_port_name: str = "o1",
    gc_rotation: int = -90,
    layer_label: LayerSpec = "TEXT",
    layer_label_loopback: Optional[Tuple[int, int]] = None,
    component_name: Optional[str] = None,
    x_grating_offset: int = 0,
    optical_port_labels: Optional[Tuple[str, ...]] = None,
    get_input_label_text_loopback_function: Callable = get_input_label_text_loopback,
    get_input_label_text_function: Optional[Callable] = get_input_label_text,
    get_input_labels_function: Optional[Callable] = get_input_labels,
    select_ports: Callable = select_ports_optical,
    cross_section: CrossSectionSpec = strip,
    **kwargs,
) -> Tuple[
    List[Union[ComponentReference, Label]], List[List[ComponentReference]], List[Port]
]:
    """Returns component I/O elements for adding grating couplers with a fiber array Many components are fine with the defaults.

    Args:
        component: component spec to connect to.
        fiber_spacing: spacing between the optical fibers.
        grating_coupler: grating coupler instance, function or list of functions.
        bend: for bends.
        straight: straight.
        fanout_length: target distance between gratings and the southest component port.
            If None, automatically calculated.
        max_y0_optical: Maximum y coordinate at which the intermediate optical ports can be set.
            Usually fine to leave at None.
        with_loopback: If True, add compact loopback alignment ports.
        nlabels_loopback: number of labels of align ports (0: no labels, 1: first port, 2: both ports2)
        straight_separation: min separation between routing straights.
        straight_to_grating_spacing: from align ports.
        optical_routing_type: There are three options for optical routing.
           * ``0`` is very basic but can be more compact.
            Can also be used in combination with ``connected_port_names``.
            or to route some components which otherwise fail with type ``1``.
           * ``1`` is the standard routing.
           * ``2`` uses the optical ports as a guideline for the component's physical size
            (instead of using the actual component size).
            Useful where the component is large due to metal tracks
           * ``None: leads to an automatic decision based on size and number
           of I/O of the component.
        connected_port_names: only for type 0 optical routing.
            Can specify which ports goes to which grating assuming the gratings are ordered from left to right.
            e.g ['N0', 'W1','W0','E0','E1', 'N1' ] or [4,1,7,3]
        nb_optical_ports_lines: number of lines with I/O grating couplers. One line by default.
            WARNING: Only works properly if:
            - nb_optical_ports_lines divides the total number of ports.
            - the components have an equal number of inputs and outputs.
        force_manhattan: sometimes port linker defaults to an S-bend due to lack of space to do manhattan.
            Force manhattan offsets all the ports to replace the s-bend by a straight link.
            This fails if multiple ports have the same issue.
        excluded_ports: ports excluded from routing.
        grating_indices: allows to fine skip some grating slots.
            e.g [0,1,4,5] will put two gratings separated by the pitch.
            Then there will be two empty grating slots, and after that an additional two gratings.
        route_filter: straight and bend factories
        gc_port_name: grating_coupler port name, where to route straights.
        gc_rotation: grating_coupler rotation (deg).
        layer_label: for TM labels.
        component_name: name of component.
        x_grating_offset: x offset.
        optical_port_labels: port labels to route_to_fiber_array.
        get_input_label_text_loopback_function: function to get input labels for grating couplers.
        get_input_label_text_function: for the label.
        get_input_labels_function: for the label.
        select_ports: function to select ports for which to add grating couplers.
        kwargs: cross_section settings.

    Returns:
        elements: list of references.
        io_grating_lines: grating lines.
        list of ports: to connect to.

    """
    fiber_spacing = gf.get_constant(fiber_spacing)
    cross_section = x = gf.get_cross_section(cross_section, **kwargs)
    radius = x.radius

    assert isinstance(
        radius, (int, float)
    ), f"radius = {radius} {type (radius)} needs to be int or float"

    component_name = component_name or component.name
    excluded_ports = excluded_ports or []
    if optical_port_labels is None:
        optical_ports = list(select_ports(component.ports).values())
    else:
        optical_ports = [component.ports[lbl] for lbl in optical_port_labels]

    optical_ports = [p for p in optical_ports if p.name not in excluded_ports]
    N = len(optical_ports)
    # optical_ports_labels = [p.name for p in optical_ports]
    # print(optical_ports_labels)
    if N == 0:
        return [], [], 0

    elements = []

    # grating_coupler can either be a component/function
    # or a list of components/functions

    if isinstance(grating_coupler, list):
        grating_couplers = [gf.call_if_func(g) for g in grating_coupler]
        grating_coupler = grating_couplers[0]
    else:
        grating_coupler = gf.call_if_func(grating_coupler)
        grating_couplers = [grating_coupler] * N

    if gc_port_name not in grating_coupler.ports:
        raise ValueError(
            f"{gc_port_name!r} not in {list(grating_coupler.ports.keys())}"
        )

    # Now:
    # - grating_coupler is a single grating coupler
    # - grating_couplers is a list of grating couplers
    # Define the route filter to apply to connection methods

    bend90 = gf.get_component(bend, cross_section=x)

    # `delta_gr_min` Used to avoid crossing between straights in special cases
    # This could happen when abs(x_port - x_grating) <= 2 * radius

    dy = bend90.info["dy"]
    delta_gr_min = 2 * dy + 1

    # Get the center along x axis
    x_c = round(sum(p.x for p in optical_ports) / N, 1)
    y_min = component.ymin  # min([p.y for p in optical_ports])

    # Sort the list of optical ports:
    direction_ports = direction_ports_from_list_ports(optical_ports)
    sep = straight_separation

    K = len(optical_ports)
    K = K + 1 if K % 2 else K

    # Set routing type if not specified
    pxs = [p.x for p in optical_ports]
    is_big_component = (
        (K > 2)
        or (max(pxs) - min(pxs) > fiber_spacing - delta_gr_min)
        or (component.xsize > fiber_spacing)
    )
    if optical_routing_type is None:
        optical_routing_type = 1 if is_big_component else 0
    # choose the default length if the default fanout distance is not set

    def has_p(side):
        return len(direction_ports[side]) > 0

    list_ew_ports_on_sides = [has_p(side) for side in ["E", "W"]]
    list_ns_ports_on_sides = [has_p(side) for side in ["N", "S"]]

    has_ew_ports = any(list_ew_ports_on_sides)
    has_ns_ports = any(list_ns_ports_on_sides)

    is_one_sided_horizontal = False
    for side1, side2 in [("E", "W"), ("W", "E")]:
        if len(direction_ports[side1]) >= 2 and all(
            len(direction_ports[side]) == 0 for side in ["N", "S", side2]
        ):
            is_one_sided_horizontal = True

    # Compute fanout length if not specified
    if fanout_length is None:
        fanout_length = dy + 1.0
        # We need 3 bends in that case to connect the most bottom port to the
        # grating couplers
        if has_ew_ports and is_big_component:
            # print('big')
            fanout_length = max(fanout_length, 3 * dy + 1.0)

        if has_ns_ports or is_one_sided_horizontal:
            # print('one sided')
            fanout_length = max(fanout_length, 2 * dy + 1.0)

        if has_ew_ports and not is_big_component:
            # print('ew_ports')
            fanout_length = max(fanout_length, dy + 1.0)

    fanout_length += dy

    # use x for grating coupler since we rotate it
    y0_optical = y_min - fanout_length - grating_coupler.ports[gc_port_name].x
    y0_optical += -K / 2 * sep
    y0_optical = round(y0_optical, 1)

    if max_y0_optical is not None:
        y0_optical = round(min(max_y0_optical, y0_optical), 1)

    # - First connect half of the north ports going from middle of list
    # down to first elements
    # - then connect west ports (top to bottom)
    # - then connect south ports (left to right)
    # - then east ports (bottom to top)
    # - then second half of the north ports (right to left)

    ports = []
    north_ports = direction_ports["N"]
    north_start = north_ports[: len(north_ports) // 2]
    north_finish = north_ports[len(north_ports) // 2 :]

    west_ports = direction_ports["W"]
    west_ports.reverse()
    east_ports = direction_ports["E"]
    south_ports = direction_ports["S"]
    north_finish.reverse()  # Sort right to left
    north_start.reverse()  # Sort right to left
    ordered_ports = north_start + west_ports + south_ports + east_ports + north_finish

    nb_ports_per_line = N // nb_optical_ports_lines
    grating_coupler_si = grating_coupler.size_info
    y_gr_gap = (K / nb_optical_ports_lines + 1) * sep
    gr_coupler_y_sep = grating_coupler_si.height + y_gr_gap + dy

    offset = (nb_ports_per_line - 1) * fiber_spacing / 2 - x_grating_offset
    io_gratings_lines = []  # [[gr11, gr12, gr13...], [gr21, gr22, gr23...] ...]

    if grating_indices is None:
        grating_indices = list(range(nb_ports_per_line))
    else:
        assert len(grating_indices) == nb_ports_per_line

    for j in range(nb_optical_ports_lines):
        io_gratings = [
            gc.ref(
                position=(
                    x_c - offset + i * fiber_spacing,
                    y0_optical - j * gr_coupler_y_sep,
                ),
                rotation=gc_rotation,
                port_id=gc_port_name,
            )
            for i, gc in zip(grating_indices, grating_couplers)
        ]

        io_gratings_lines += [io_gratings[:]]
        ports += [grating.ports[gc_port_name] for grating in io_gratings]

    if optical_routing_type == 0:
        """Basic optical routing, typically fine for small components No
        heuristic to avoid collisions between connectors.

        If specified ports to connect in a specific order (i.e if
        connected_port_names is not None and not empty) then grab these
        ports

        """
        if connected_port_names:
            ordered_ports = [component.ports[i] for i in connected_port_names]

        for io_gratings in io_gratings_lines:
            for i in range(N):
                p0 = io_gratings[i].ports[gc_port_name]
                p1 = ordered_ports[i]
                waypoints = generate_manhattan_waypoints(
                    input_port=p0,
                    output_port=p1,
                    bend=bend90,
                    straight=straight,
                    cross_section=cross_section,
                )
                route = route_filter(
                    waypoints=waypoints,
                    bend=bend90,
                    straight=straight,
                    cross_section=cross_section,
                )
                elements.extend(route.references)

    elif optical_routing_type in [1, 2]:
        route = route_south(
            component=component,
            optical_routing_type=optical_routing_type,
            excluded_ports=excluded_ports,
            straight_separation=straight_separation,
            io_gratings_lines=io_gratings_lines,
            gc_port_name=gc_port_name,
            bend=bend90,
            straight=straight,
            taper=taper,
            select_ports=select_ports,
            cross_section=cross_section,
        )
        elems = route.references
        to_route = route.ports
        elements.extend(elems)

        if force_manhattan:
            """1) find the min x_distance between each grating port and each
            component port.

            2) If abs(min distance) < 2* bend radius     then offset
            io_gratings by -min_distance

            """
            min_dist = 2 * dy + 10.0
            min_dist_threshold = 2 * dy + 1.0
            for io_gratings in io_gratings_lines:
                for gr in io_gratings:
                    for p in to_route:
                        dist = gr.x - p.x
                        if abs(dist) < abs(min_dist):
                            min_dist = dist
                if abs(min_dist) < min_dist_threshold:
                    for gr in io_gratings:
                        gr.movex(-min_dist)

        # If the array of gratings is too close, adjust its location
        gc_ports_tmp = []
        for io_gratings in io_gratings_lines:
            gc_ports_tmp += [gc.ports[gc_port_name] for gc in io_gratings]
        min_y = get_min_spacing(to_route, gc_ports_tmp, sep=sep, radius=dy)
        delta_y = abs(to_route[0].y - gc_ports_tmp[0].y)

        if min_y > delta_y:
            for io_gratings in io_gratings_lines:
                for gr in io_gratings:
                    gr.translate(0, delta_y - min_y)

        # If we add align ports, we need enough space for the bends
        end_straight_offset = straight_separation + 5 if with_loopback else x.min_length
        if len(io_gratings_lines) == 1:
            io_gratings = io_gratings_lines[0]
            gc_ports = [gc.ports[gc_port_name] for gc in io_gratings]
            routes = get_bundle(
                ports1=to_route,
                ports2=gc_ports,
                separation=sep,
                end_straight_length=end_straight_offset,
                straight=straight,
                bend=bend90,
                cross_section=cross_section,
            )
            elements.extend([route.references for route in routes])

        else:
            for io_gratings in io_gratings_lines:
                gc_ports = [gc.ports[gc_port_name] for gc in io_gratings]
                nb_gc_ports = len(io_gratings)
                nb_ports_to_route = len(to_route)
                n0 = nb_ports_to_route / 2
                dn = nb_gc_ports / 2
                routes = get_bundle(
                    ports1=to_route[n0 - dn : n0 + dn],
                    ports2=gc_ports,
                    separation=sep,
                    end_straight_length=end_straight_offset,
                    bend=bend90,
                    straight=straight,
                    cross_section=cross_section,
                )
                elements.extend([route.references for route in routes])
                del to_route[n0 - dn : n0 + dn]

    if with_loopback:
        gca1, gca2 = (
            grating_coupler.ref(
                position=(
                    x_c - offset + ii * fiber_spacing,
                    io_gratings_lines[-1][0].ports[gc_port_name].y,
                ),
                rotation=gc_rotation,
                port_id=gc_port_name,
            )
            for ii in [grating_indices[0] - 1, grating_indices[-1] + 1]
        )
        port0 = gca1.ports[gc_port_name]
        port1 = gca2.ports[gc_port_name]
        ports.append(port0)
        ports.append(port1)

        p0 = port0.center
        p1 = port1.center

        dy = bend90.info["dy"]
        dx = max(2 * dy, fiber_spacing / 2)

        gc_east = max(gci.size_info.east for gci in grating_couplers)
        y_bot_align_route = gc_east + straight_to_grating_spacing

        points = [
            p0,
            p0 + (0, dy),
            p0 + (dx, dy),
            p0 + (dx, -y_bot_align_route),
            p1 + (-dx, -y_bot_align_route),
            p1 + (-dx, dy),
            p1 + (0, dy),
            p1,
        ]
        io_gratings_lines += [[gca1], [gca2]]
        # elements.extend([gca1, gca2])

        route = round_corners(
            points=points,
            straight=straight,
            bend=bend90,
            cross_section=cross_section,
        )
        elements.extend(route.references)
        if nlabels_loopback == 1:
            io_gratings_loopback = [gca1]
            ordered_ports_loopback = [port0]
        if nlabels_loopback == 2:
            io_gratings_loopback = [gca1, gca2]
            ordered_ports_loopback = [port0, port1]
        elif nlabels_loopback > 2:
            raise ValueError(
                f"Invalid nlabels_loopback = {nlabels_loopback}, "
                "valid (0: no labels, 1: first port, 2: both ports2)"
            )
        if (
            nlabels_loopback > 0
            and get_input_labels_function
            and get_input_label_text_function
        ):
            elements.extend(
                get_input_labels_function(
                    io_gratings=io_gratings_loopback,
                    ordered_ports=ordered_ports_loopback,
                    component_name=component_name,
                    layer_label=layer_label_loopback or layer_label,
                    gc_port_name=gc_port_name,
                    get_input_label_text_function=get_input_label_text_loopback_function,
                )
            )

    if get_input_labels_function and get_input_label_text_function:
        elements.extend(
            get_input_labels_function(
                io_gratings=io_gratings,
                ordered_ports=ordered_ports,
                component_name=component_name,
                layer_label=layer_label,
                gc_port_name=gc_port_name,
                get_input_label_text_function=get_input_label_text_function,
            )
        )

    return elements, io_gratings_lines, ports


def demo() -> None:
    gcte = gf.components.grating_coupler_te
    gctm = gf.components.grating_coupler_tm

    c = gf.components.straight(length=500)
    c = gf.components.mmi2x2()

    elements, gc, _ = route_fiber_array(
        component=c,
        grating_coupler=[gcte, gctm, gcte, gctm],
        with_loopback=True,
        optical_routing_type=2,
        # bend=gf.components.bend_euler,
        bend=gf.components.bend_circular,
        radius=20,
        # force_manhattan=True
    )
    for e in elements:
        # if isinstance(e, list):
        # print(len(e))
        # print(e)
        c.add(e)
    for e in gc:
        c.add(e)
    c.show(show_ports=True)


if __name__ == "__main__":
    # layer = (2, 0)
    # c = gf.components.straight(layer=layer)
    # gc = gf.components.grating_coupler_elliptical_te(layer=layer, taper_length=30)
    # gc.xmin = -20
    # elements, gc, _ = route_fiber_array(
    #     component=c,
    #     grating_coupler=gc,
    #     cladding_offset=6,
    #     nlabels_loopback=1,
    #     layer=layer,
    # )
    # # c = p.ring_single()
    # # c = p.add_fiber_array(c, optical_routing_type=1, auto_widen=False)
    # for e in elements:
    #     # if isinstance(e, list):
    #     # print(len(e))
    #     # print(e)
    #     c.add(e)
    # for e in gc:
    #     c.add(e)

    c = gf.Component()

    layer = (1, 0)
    ci = gf.components.straight(layer=layer)
    ci = gf.components.mmi2x2()
    ci = gf.components.straight_heater_metal()
    gc = gf.components.grating_coupler_elliptical_te(layer=layer, taper_length=30)
    gc.xmin = -20
    elements, gc, ports = route_fiber_array(
        component=ci,
        grating_coupler=gc,
        nlabels_loopback=1,
        layer=layer,
        # optical_routing_type=2,
        # fanout_length=20,
        # get_input_label_text_function=None,
    )
    # c = p.ring_single()
    # c = p.add_fiber_array(c, optical_routing_type=1, auto_widen=False)
    c << ci
    for e in elements:
        # if isinstance(e, list):
        # print(len(e))
        # print(e)
        c.add(e)
    for e in gc:
        c.add(e)
    c.show(show_ports=True)



"""Bends with grating couplers inside the spiral.

maybe: need to add grating coupler loopback as well
"""
from typing import Optional, Tuple

import numpy as np
from numpy import float64

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.routing.manhattan import round_corners
from gdsfactory.types import ComponentSpec, CrossSectionSpec


def get_bend_port_distances(bend: Component) -> Tuple[float64, float64]:
    """Returns distance between bend ports."""
    p0, p1 = bend.ports.values()
    return abs(p0.x - p1.x), abs(p0.y - p1.y)


@cell
def spiral_external_io(
    N: int = 6,
    x_inner_length_cutback: float = 300.0,
    x_inner_offset: float = 0.0,
    y_straight_inner_top: float = 0.0,
    xspacing: float = 3.0,
    yspacing: float = 3.0,
    bend: ComponentSpec = bend_euler,
    length: Optional[float] = None,
    cross_section: CrossSectionSpec = "strip",
    **kwargs
) -> Component:
    """Returns spiral with input and output ports outside the spiral.

    Args:
        N: number of loops.
        x_inner_length_cutback: x inner length.
        x_inner_offset: x inner offset.
        y_straight_inner_top: y straight inner top.
        xspacing: center to center x-spacing.
        yspacing: center to center y-spacing.
        bend: function.
        length: length in um, it is the approximates total length.
        cross_section: spec.
        kwargs: cross_section settings.
    """
    if length:
        x_inner_length_cutback = length / (4 * (N - 1))

    y_straight_inner_top += 5
    x_inner_length_cutback += x_inner_offset
    _bend180 = bend(angle=180, cross_section=cross_section, **kwargs)
    _bend90 = bend(angle=90, cross_section=cross_section, **kwargs)

    bend_radius = _bend90.info["radius"]
    rx, ry = get_bend_port_distances(_bend90)
    _, rx180 = get_bend_port_distances(_bend180)  # rx180, second arg since we rotate

    component = Component()
    inner_loop_spacing = 2 * bend_radius + 5.0

    # Create manhattan path going from west grating to westest port of bend 180
    x_inner_length = x_inner_length_cutback + 5.0 + xspacing

    y_inner_bend = y_straight_inner_top - bend_radius - 5.0
    x_inner_loop = x_inner_length - 5.0
    p1 = (x_inner_loop, y_inner_bend)
    p2 = (x_inner_loop + inner_loop_spacing, y_inner_bend)

    _pt = np.array(p1)
    pts_w = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i + 1) * yspacing
        x2 = inner_loop_spacing + 2 * rx + x_inner_length + (2 * i + 1) * xspacing
        y3 = -ry - (2 * i + 2) * yspacing
        x4 = -(2 * i + 1) * xspacing
        if i == N - 1:
            x4 = x4 - rx180 + xspacing

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_w += [_pt1, _pt2, _pt3, _pt4, _pt5]

    pts_w = pts_w[:-2]

    # Create manhattan path going from east grating to eastest port of bend 180
    _pt = np.array(p2)
    pts_e = [_pt]

    for i in range(N):
        y1 = y_straight_inner_top + ry + (2 * i) * yspacing
        x2 = inner_loop_spacing + 2 * rx + x_inner_length + 2 * i * xspacing
        y3 = -ry - (2 * i + 1) * yspacing
        x4 = -2 * i * xspacing

        _pt1 = np.array([_pt[0], y1])
        _pt2 = np.array([x2, _pt1[1]])
        _pt3 = np.array([_pt2[0], y3])
        _pt4 = np.array([x4, _pt3[1]])
        _pt5 = np.array([_pt4[0], 0])
        _pt = _pt5

        pts_e += [_pt1, _pt2, _pt3, _pt4, _pt5]

    pts_e = pts_e[:-2]

    # Join the two bits of paths and extrude the spiral geometry
    route = round_corners(
        pts_w[::-1] + pts_e,
        bend=bend,
        cross_section=cross_section,
        **kwargs,
    )

    component.add(route.references)
    component.add_port("o2", port=route.ports[0])
    component.add_port("o1", port=route.ports[1])

    length = route.length
    component.info["length"] = length
    return component


if __name__ == "__main__":
    c = spiral_external_io(auto_widen=True, width_wide=2.0, length=10e3, N=15)
    # print(c.info['length'])
    # print(c.info['length'] / 1e4, "cm")
    c.show(show_ports=True)



import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.grating_coupler_elliptical import grating_tooth_points
from gdsfactory.geometry.functions import DEG2RAD
from gdsfactory.types import CrossSectionSpec, LayerSpec, Optional


@gf.cell
def grating_coupler_elliptical_trenches(
    polarization: str = "te",
    fiber_marker_width: float = 11.0,
    fiber_marker_layer: Optional[LayerSpec] = "TE",
    taper_length: float = 16.6,
    taper_angle: float = 30.0,
    trenches_extra_angle: float = 9.0,
    wavelength: float = 1.53,
    fiber_angle: float = 15.0,
    grating_line_width: float = 0.343,
    neff: float = 2.638,  # tooth effective index
    ncladding: float = 1.443,  # cladding index
    layer_trench: LayerSpec = "SLAB150",
    p_start: int = 26,
    n_periods: int = 30,
    end_straight_length: float = 0.2,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    r"""Returns Grating coupler with defined trenches.

    Some foundries define the grating coupler by a shallow etch step (trenches)
    Others define the slab that they keep (see grating_coupler_elliptical)

    Args:
        polarization: 'te' or 'tm'.
        taper_length: taper length from straight I/O.
        taper_angle: grating flare angle.
        wavelength: grating transmission central wavelength.
        fiber_angle: fibre polish angle in degrees.
        grating_line_width: of the 220 ridge.
        neff: tooth effective index.
        ncladding: cladding index.
        layer_trench: for the trench.
        p_start: first tooth.
        n_periods: number of grating teeth.
        end_straight_length: at the end of straight.
        kwargs: cross_section settings.


    .. code::

                      fiber

                   /  /  /  /
                  /  /  /  /
                _|-|_|-|_|-|___
        WG  o1  ______________|

    """
    xs = gf.get_cross_section(cross_section, **kwargs)
    wg_width = xs.width
    layer = xs.layer

    # Compute some ellipse parameters
    sthc = np.sin(fiber_angle * DEG2RAD)
    d = neff**2 - ncladding**2 * sthc**2
    a1 = wavelength * neff / d
    b1 = wavelength / np.sqrt(d)
    x1 = wavelength * ncladding * sthc / d

    a1 = round(a1, 3)
    b1 = round(b1, 3)
    x1 = round(x1, 3)

    period = float(a1 + x1)
    trench_line_width = period - grating_line_width

    c = gf.Component()

    # Make each grating line
    for p in range(p_start, p_start + n_periods + 1):
        pts = grating_tooth_points(
            p * a1,
            p * b1,
            p * x1,
            width=trench_line_width,
            taper_angle=taper_angle + trenches_extra_angle,
        )
        c.add_polygon(pts, layer_trench)

    # Make the taper
    p_taper = p_start - 1
    p_taper_eff = p_taper
    a_taper = a1 * p_taper_eff
    # b_taper = b1 * p_taper_eff
    x_taper = x1 * p_taper_eff
    x_output = a_taper + x_taper - taper_length + grating_line_width / 2

    xmax = x_output + taper_length + n_periods * period + 3
    y = wg_width / 2 + np.tan(taper_angle / 2 * np.pi / 180) * xmax
    pts = [
        (x_output, -wg_width / 2),
        (x_output, wg_width / 2),
        (xmax, y),
        (xmax + end_straight_length, y),
        (xmax + end_straight_length, -y),
        (xmax, -y),
    ]
    c.add_polygon(pts, layer)

    x = np.round(taper_length + period * n_periods / 2, 3)
    if fiber_marker_layer:
        circle = gf.components.circle(
            radius=fiber_marker_width / 2, layer=fiber_marker_layer
        )
        circle_ref = c.add_ref(circle)
        circle_ref.movex(x)
    name = f"vertical_{polarization.lower()}"
    c.add_port(
        name=name,
        center=(x, 0),
        width=fiber_marker_width,
        orientation=0,
        layer=fiber_marker_layer,
        port_type=name,
    )

    c.add_port(
        name="o1",
        center=(x_output, 0),
        width=wg_width,
        orientation=180,
        layer=layer,
        cross_section=xs,
    )
    c.info["period"] = period
    c.info["polarization"] = polarization
    c.info["wavelength"] = wavelength
    if xs.add_bbox:
        c = xs.add_bbox(c)
    if xs.add_pins:
        c = xs.add_pins(c)
    return c


grating_coupler_te = gf.partial(
    grating_coupler_elliptical_trenches, polarization="te", taper_angle=35
)

grating_coupler_tm = gf.partial(
    grating_coupler_elliptical_trenches,
    polarization="tm",
    neff=1.8,
    grating_line_width=0.6,
    fiber_marker_layer="TM",
)


if __name__ == "__main__":
    # c = grating_coupler_elliptical_trenches(polarization="TE")
    # print(c.polarization)
    # c = grating_coupler_te(end_straight_length=10, fiber_marker_layer=None)
    # c = grating_coupler_tm()
    # print(c.ports.keys())
    c = gf.routing.add_fiber_array(grating_coupler=grating_coupler_elliptical_trenches)
    c.show(show_ports=True)



"""[[-0.09051371-0.20581339j  0.00704022+0.1328474j \
    0.03733851+0.4879802j ]."""

from pathlib import PosixPath
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
from simphony import Model
from simphony.tools import freq2wl, interpolate, wl2freq

import gdsfactory as gf
from gdsfactory.simulation.lumerical.read import read_sparameters_file


class SimphonyFromFile(Model):
    """Take an s-parameter file path and return a Simphony Model from it."""

    def __init__(
        self,
        numports: Optional[int] = None,
        pins: Optional[Tuple[str, ...]] = None,
        name: str = "model",
    ) -> None:
        """Take an s-parameter file path and return a Simphony Model from it.

        Uses numports if passed, else uses pins.

        Args:
            numports: number of ports.
            pins: list of port names.
            name: optional model name.
        """
        if numports:
            __class__.pin_count = numports
        elif pins:
            __class__.pin_count = len(pins)
        else:
            raise ValueError("Either numports or pin must be defined.")

        super().__init__(name)

        if pins:
            self.rename_pins(*pins)

    def model_from_filepath(
        self,
        filepath: PosixPath,
    ) -> Model:
        """Returns simphony Model from lumerical .DAT sparameters.

        Args:
            filepath: path to Sparameters in Lumerical interconnect format.

        """
        numports = self.pin_count
        pin_names, self.f, self.s = read_sparameters_file(
            filepath=filepath, numports=numports
        )
        self.wavelengths = freq2wl(self.f)

        self.rename_pins(*pin_names)

        return self

    def model_from_csv(
        self,
        filepath: Union[str, PosixPath, pd.DataFrame],
        xkey: str = "wavelengths",
        xunits: float = 1,
    ) -> Model:
        """Returns simphony Model from Sparameters in CSV.

        Args:
            filepath: CSV Sparameters path or pandas DataFrame.
            xkey: key for wavelengths in file.
            xunits: x units in um from the loaded file (um). 1 means 1um.

        """
        df = filepath if isinstance(filepath, pd.DataFrame) else pd.read_csv(filepath)

        keys = list(df.keys())

        if xkey not in df:
            raise ValueError(f"{xkey!r} not in {keys}")

        self.wavelengths = df[xkey].values * 1e-6 * xunits

        numrows = len(self.wavelengths)
        numports = self.pin_count
        self.s = np.zeros((numrows, numports, numports), dtype="complex128")

        for i in range(numports):
            for j in range(numports):
                self.s[:, i, j] = df[f"s{i+1}{j+1}m"] * np.exp(1j * df[f"s{i+1}{j+1}a"])

        self.f = wl2freq(self.wavelengths)
        self.freq_range = (self.f[0], self.f[-1])

        return self

    def s_parameters(self, freqs: "np.array") -> "np.ndarray":
        return interpolate(freqs, self.f, self.s)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    from gdsfactory.simulation.simphony.plot_model import plot_model

    filepath = gf.CONFIG["sparameters"] / "mmi1x2" / "mmi1x2_si220n.dat"
    numports = 3
    c = SimphonyFromFile(numports=numports).model_from_filepath(filepath=filepath)

    filepath_csv = gf.CONFIG["sparameters"] / "mmi1x2" / "mmi1x2_si220n.csv"
    # filepath_csv = "/home/jmatres/ubc/sparameters/ebeam_y_1550_20634f71.csv"
    # df = pd.read_csv(filepath_csv)
    c = SimphonyFromFile(pins=("o1", "o2", "o3")).model_from_csv(filepath_csv)
    plot_model(c, pin_in="o1")
    plt.show()

    # wav = np.linspace(1520, 1570, 1024) * 1e-9
    # f = speed_of_light / wav
    # s = c.s_parameters(freqs=f)
    # wav = c.wavelengths
    # s = c.s
    # plt.plot(wav * 1e9, np.abs(s[:, 1] ** 2))



import pytest

import gdsfactory as gf
from gdsfactory.routing.manhattan import RouteWarning


def test_route_error():
    """Ensures that an impossible route raises RouteWarning."""
    c = gf.Component("test_route_error")
    w = gf.components.straight()
    left = c << w
    right = c << w
    right.move((100, 80))

    p1 = left.ports["o2"]
    p2 = right.ports["o2"]

    with pytest.warns(RouteWarning):
        route = gf.routing.get_route_from_steps(
            port1=p2,
            port2=p1,
            steps=[
                {"x": 20, "y": 0},
                {"x": 20, "y": 20},
                {"x": 120, "y": 20},
                {"x": 120, "y": 80},
            ],
        )
        c.add(route.references)
    return c


if __name__ == "__main__":
    c = test_route_error()
    c.show(show_ports=True)



import time

import numpy as np
import tidy3d as td
from omegaconf import OmegaConf

import gdsfactory as gf
from gdsfactory.config import logger
from gdsfactory.serialization import clean_value_json
from gdsfactory.simulation import port_symmetries
from gdsfactory.simulation.get_sparameters_path import (
    get_sparameters_path_tidy3d as get_sparameters_path,
)
from gdsfactory.simulation.gtidy3d.get_results import _executor, get_results
from gdsfactory.simulation.gtidy3d.get_simulation import get_simulation, plot_simulation
from gdsfactory.types import (
    Any,
    ComponentSpec,
    Dict,
    List,
    Optional,
    PathType,
    Port,
    PortSymmetries,
    Tuple,
)


def parse_port_eigenmode_coeff(
    port_name: str, ports: Dict[str, Port], sim_data: td.SimulationData
) -> Tuple[np.ndarray]:
    """Given a port and eigenmode coefficient result, returns the coefficients \
    relative to whether the wavevector is entering or exiting simulation.

    Args:
        port_name: port name.
        ports: component_ref.ports.
        sim_data: simulation data.
    """
    # Direction of port (pointing away from the simulation)
    # Figure out if that is exiting the simulation or not
    # depending on the port orientation (assuming it's near PMLs)

    orientation = ports[port_name].orientation

    if orientation in [0, 90]:  # east
        direction_inp = "-"
        direction_out = "+"
    elif orientation in [180, 270]:  # west
        direction_inp = "+"
        direction_out = "-"
    else:
        raise ValueError(
            "Port orientation = {orientation} is not 0, 90, 180, or 270 degrees"
        )

    coeff_inp = sim_data.monitor_data[port_name].amps.sel(direction=direction_inp)
    coeff_out = sim_data.monitor_data[port_name].amps.sel(direction=direction_out)
    return coeff_inp.values.flatten(), coeff_out.values.flatten()


def get_wavelengths(port_name: str, sim_data: td.SimulationData) -> np.ndarray:
    coeff_inp = sim_data.monitor_data[port_name].amps.sel(direction="+")
    freqs = coeff_inp.f
    return td.constants.C_0 / freqs.values


def write_sparameters(
    component: ComponentSpec,
    port_symmetries: Optional[PortSymmetries] = None,
    port_source_names: Optional[List[str]] = None,
    dirpath: Optional[PathType] = None,
    run: bool = True,
    overwrite: bool = False,
    **kwargs,
) -> np.ndarray:
    """Get full sparameter matrix from a gdsfactory Component.

    Simulates each time using a different input port (by default, all of them)
    unless you specify port_symmetries.

    port_symmetries = {"o1":
            {
                "s11": ["s22","s33","s44"],
                "s21": ["s21","s34","s43"],
                "s31": ["s13","s24","s42"],
                "s41": ["s14","s23","s32"],
            }
        }
    - Only simulations using the outer key port names will be run
    - The associated value is another dict whose keys are the S-parameters computed
        when this source is active
    - The values of this inner Dict are lists of s-parameters whose values are copied

    Args:
        component: to simulate.
        port_source_names: list of ports to excite. Defaults to all.
        port_symmetries: Dict to specify port symmetries, to save number of simulations
        dirpath: directory to store sparameters in npz.
            Defaults to active Pdk.sparameters_path.
        run: runs simulation, if False, only plots simulation.
        overwrite: overwrites stored Sparameter npz results.

    Keyword Args:
        port_extension: extend ports beyond the PML.
        layer_stack: contains layer to thickness, zmin and material.
            Defaults to active pdk.layer_stack.
        thickness_pml: PML thickness (um).
        xmargin: left/right distance from component to PML.
        xmargin_left: left distance from component to PML.
        xmargin_right: right distance from component to PML.
        ymargin: left/right distance from component to PML.
        ymargin_top: top distance from component to PML.
        ymargin_bot: bottom distance from component to PML.
        zmargin: thickness for cladding above and below core.
        clad_material: material for cladding.
        port_margin: margin on each side of the port.
        distance_source_to_monitors: in (um) source goes before monitors.
        wavelength_start: in (um).
        wavelength_stop: in (um).
        wavelength_points: in (um).
        plot_modes: plot source modes.
        num_modes: number of modes to plot.
        run_time_ps: make sure it's sufficient for the fields to decay.
            defaults to 10ps and counts on automatic shutoff to stop earlier if needed.
        dispersive: False uses constant refractive index materials.
            True adds wavelength depending materials.
            Dispersive materials require more computation.
        material_name_to_tidy3d_index: not dispersive materials have a constant index.
        material_name_to_tidy3d_name: dispersive materials have a wavelength
            dependent index. Maps layer_stack names with tidy3d material database names.
        is_3d: if False, does not consider Z dimension for faster simulations.
        with_all_monitors: True adds field monitor which increases results file size.
        grid_spec: defaults to automatic td.GridSpec.auto(wavelength=wavelength)
            td.GridSpec.uniform(dl=20*nm)
            td.GridSpec(
                grid_x = td.UniformGrid(dl=0.04),
                grid_y = td.AutoGrid(min_steps_per_wvl=20),
                grid_z = td.AutoGrid(min_steps_per_wvl=20),
                wavelength=wavelength,
                override_structures=[refine_box]
            )
        dilation: float = 0.0
            Dilation of the polygon in the base by shifting each edge along its
            normal outwards direction by a distance;
            a negative value corresponds to erosion.
        sidewall_angle_deg : float = 0
            Angle of the sidewall.
            ``sidewall_angle=0`` (default) specifies vertical wall,
            while ``0<sidewall_angle_deg<90`` for the base to be larger than the top.

    """
    component = gf.get_component(component)
    filepath = get_sparameters_path(
        component=component,
        dirpath=dirpath,
        **kwargs,
    )
    filepath_sim_settings = filepath.with_suffix(".yml")
    if filepath.exists() and not overwrite and run:
        logger.info(f"Simulation loaded from {filepath!r}")
        return np.load(filepath)

    port_symmetries = port_symmetries or {}
    component_ref = component.ref()
    ports = component_ref.ports
    port_names = [port.name for port in list(ports.values())]

    sims = []
    sp = {}

    port_source_names = port_source_names or port_names

    for port_name in port_source_names:
        if port_name not in port_symmetries:
            sim = get_simulation(component, port_source_name=port_name, **kwargs)
            sims.append(sim)

    if not run:
        sim = sims[0]
        plot_simulation(sim)
        return sp

    start = time.time()
    batch_data = get_results(sims, overwrite=overwrite)

    def get_sparameter(
        port_name_source: str,
        sim_data: td.SimulationData,
        port_symmetries=port_symmetries,
        **kwargs,
    ) -> np.ndarray:
        """Return Component sparameter for a particular port Index n.

        Args:
            port_name: source port name.
            sim_data: simulation data.
            port_symmetries: to save simulations.
            kwargs: simulation settings.
        """
        source_entering, source_exiting = parse_port_eigenmode_coeff(
            port_name=port_name_source, ports=component_ref.ports, sim_data=sim_data
        )

        for port_name in port_names:
            monitor_entering, monitor_exiting = parse_port_eigenmode_coeff(
                port_name=port_name, ports=ports, sim_data=sim_data
            )
            sij = monitor_exiting / source_entering
            key = f"{port_name}@0,{port_name_source}@0"
            sp[key] = sij
            sp["wavelengths"] = get_wavelengths(port_name=port_name, sim_data=sim_data)

        if bool(port_symmetries):
            for key, symmetries in port_symmetries.items():
                for sym in symmetries:
                    if key in sp:
                        sp[sym] = sp[key]

        return sp

    for port_source_name, (_sim_name, sim_data) in zip(
        port_source_names, batch_data.items()
    ):
        sp.update(get_sparameter(port_source_name, sim_data))

    end = time.time()
    np.savez_compressed(filepath, **sp)
    kwargs.update(compute_time_seconds=end - start)
    kwargs.update(compute_time_minutes=(end - start) / 60)
    filepath_sim_settings.write_text(OmegaConf.to_yaml(clean_value_json(kwargs)))
    logger.info(f"Write simulation results to {str(filepath)!r}")
    logger.info(f"Write simulation settings to {str(filepath_sim_settings)!r}")
    return sp


def write_sparameters_batch(jobs: List[Dict[str, Any]], **kwargs) -> List[np.ndarray]:
    """Returns Sparameters for a list of write_sparameters_grating_coupler kwargs \
            where it runs each simulation in parallel.

    Args:
        jobs: list of kwargs for write_sparameters_grating_coupler.
        kwargs: simulation settings.
    """
    sp = [_executor.submit(write_sparameters, **job, **kwargs) for job in jobs]
    return [spi.result() for spi in sp]


write_sparameters_1x1 = gf.partial(
    write_sparameters, port_symmetries=port_symmetries.port_symmetries_1x1
)
write_sparameters_crossing = gf.partial(
    write_sparameters, port_symmetries=port_symmetries.port_symmetries_crossing
)

write_sparameters_batch_1x1 = gf.partial(
    write_sparameters_batch, port_symmetries=port_symmetries.port_symmetries_1x1
)


if __name__ == "__main__":
    import gdsfactory as gf
    import gdsfactory.simulation as sim

    # c = gf.components.straight(length=2.1)
    c = gf.c.straight()
    c = gf.components.mmi1x2()
    sp = write_sparameters(c, is_3d=True, port_source_names=None, overwrite=False)
    sim.plot.plot_sparameters(sp)

    # t = sp.o1@0,o2@0
    # print(f"Transmission = {t}")
    # cs = [gf.c.straight(length=1.11 + i) for i in [1, 2]]
    # sps = write_sparameters_batch_1x1(cs)
